"""
Allow stepping through time using :mod:`xarray`'s offset functionality

Provides similar functionality to https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects
"""
from typing import Iterable

import cftime
from xarray.coding import cftime_offsets
from xarray.coding.cftime_offsets import to_offset  # noqa: F401, E402


def generate_range(
    start: cftime.datetime,
    end: cftime.datetime,
    offset: cftime_offsets.BaseCFTimeOffset,
) -> Iterable[cftime.datetime]:
    """
    Generate a range of datetime objects between start and end, using offset to
    determine the steps.

    The range will extend both ends of the span to the next valid timestep, see
    examples.

    Parameters
    ----------
    start: cftime.datetime
        Starting datetime from which to generate the range (noting roll backward
        mentioned above and illustrated in the examples).

    end: cftime.datetime
        Last datetime from which to generate the range (noting roll forward mentioned
        above and illustrated in the examples).

    offset:
        Offset object for determining the timesteps.

    Yields
    ------
    :obj:`cftime.datetime`
        Next datetime in the range

    Raises
    ------
    ValueError
        Offset does not result in increasing :class:`cftime.datetime`'s

    Examples
    --------
    The range is extended at either end to the nearest timestep. In the example below,
    the first timestep is rolled back to 1st Jan 2001 whilst the last is extended to 1st
    Jan 2006.

    >>> import datetime as dt
    >>> from pprint import pprint
    >>> from scmdata.offsets import to_offset, generate_range
    >>> g = generate_range(
    ...     dt.datetime(2001, 4, 1),
    ...     dt.datetime(2005, 6, 3),
    ...     to_offset("AS"),
    ... )

    >>> pprint([d for d in g])
    [cftime.datetime(2001, 1, 1, 0, 0),
     cftime.datetime(2002, 1, 1, 0, 0),
     cftime.datetime(2003, 1, 1, 0, 0),
     cftime.datetime(2004, 1, 1, 0, 0),
     cftime.datetime(2005, 1, 1, 0, 0),
     cftime.datetime(2006, 1, 1, 0, 0)]

    In this example the first timestep is rolled back to 31st Dec 2000 whilst the last
    is extended to 31st Dec 2005.

    >>> g = generate_range(
    ...     dt.datetime(2001, 4, 1),
    ...     dt.datetime(2005, 6, 3),
    ...     to_offset("A"),
    ... )
    >>> pprint([d for d in g])
    [cftime.datetime(2000, 12, 31, 0, 0),
     cftime.datetime(2001, 12, 31, 0, 0),
     cftime.datetime(2002, 12, 31, 0, 0),
     cftime.datetime(2003, 12, 31, 0, 0),
     cftime.datetime(2004, 12, 31, 0, 0),
     cftime.datetime(2005, 12, 31, 0, 0)]

    In this example the first timestep is already on the offset so stays there, the last
    timestep is to 1st Sep 2005.

    >>> g = generate_range(
    ...     dt.datetime(2001, 4, 1),
    ...     dt.datetime(2005, 6, 3),
    ...     to_offset("QS"),
    ... )
    >>> pprint([d for d in g])
    [cftime.datetime(2001, 4, 1, 0, 0),
     cftime.datetime(2001, 7, 1, 0, 0),
     cftime.datetime(2001, 10, 1, 0, 0),
     cftime.datetime(2002, 1, 1, 0, 0),
     cftime.datetime(2002, 4, 1, 0, 0),
     cftime.datetime(2002, 7, 1, 0, 0),
     cftime.datetime(2002, 10, 1, 0, 0),
     cftime.datetime(2003, 1, 1, 0, 0),
     cftime.datetime(2003, 4, 1, 0, 0),
     cftime.datetime(2003, 7, 1, 0, 0),
     cftime.datetime(2003, 10, 1, 0, 0),
     cftime.datetime(2004, 1, 1, 0, 0),
     cftime.datetime(2004, 4, 1, 0, 0),
     cftime.datetime(2004, 7, 1, 0, 0),
     cftime.datetime(2004, 10, 1, 0, 0),
     cftime.datetime(2005, 1, 1, 0, 0),
     cftime.datetime(2005, 4, 1, 0, 0),
     cftime.datetime(2005, 7, 1, 0, 0)]
    """
    # Uses the Gregorian calendar - allows for adding/subtracting datetime.timedelta in range calc
    start_cf = cftime.DatetimeGregorian(*start.timetuple()[:6])
    end_cf = cftime.DatetimeGregorian(*end.timetuple()[:6])

    res = cftime_offsets.cftime_range(
        offset.rollback(start_cf), offset.rollforward(end_cf), freq=offset
    )

    return [cftime.datetime(*dt.timetuple()[:6]) for dt in res]



"""
Testing utilities
"""

import numpy as np
import numpy.testing as npt
import packaging.version
import pandas as pd
import pandas.testing as pdt


def _check_pandas_less_110():
    return packaging.version.parse(pd.__version__) < packaging.version.Version("1.1.0")


def _assert_frame_equal(left, right, **kwargs):
    if _check_pandas_less_110():
        kwargs.pop("rtol", None)
        kwargs.pop("atol", None)

    pdt.assert_frame_equal(left, right, **kwargs)


def assert_scmdf_almost_equal(
    left, right, allow_unordered=False, check_ts_names=True, rtol=1e-5, atol=1e-8
):
    """
    Check that left and right :obj:`ScmRun` are equal.

    Parameters
    ----------
    left : :obj:`ScmRun`

    right : :obj:`ScmRun`

    allow_unordered : bool
        If true, the order of the timeseries is not checked

    check_ts_names : bool
        If true, check that the meta names are the same

    rtol : float
        Relative tolerance on numeric comparisons

    atol : float
        Absolute tolerance on numeric comparisons

    Raises
    ------
    AssertionError
        ``left`` and ``right`` are not equal
    """
    # Check that the meta data is close
    if allow_unordered or not check_ts_names:

        # Checks that all the timeseries are named the same
        if check_ts_names:
            df1_index = np.argsort(left.meta.index)
            df2_index = np.argsort(right.meta.index)
            _assert_frame_equal(left.meta, right.meta, check_like=True)
            npt.assert_allclose(
                left.values[df1_index], right.values[df2_index], rtol=rtol, atol=atol
            )

        else:
            # ignore differing meta index labels
            # instead sort by meta values then check equality
            left_sorted = left.timeseries().sort_index()
            right_ts = right.timeseries()
            # check metadata columns are same set
            if set(left_sorted.index.names) != set(right_ts.index.names):
                raise AssertionError(
                    "{} != {}".format(
                        set(left_sorted.index.names), set(right_ts.index.names)
                    )
                )

            right_sorted = right.timeseries(left_sorted.index.names).sort_index()
            # this checks both the index (i.e. sorted meta) and values are the same
            _assert_frame_equal(left_sorted, right_sorted, rtol=rtol, atol=atol)

    else:
        _assert_frame_equal(left.meta, right.meta)
        npt.assert_allclose(left.values, right.values, rtol=rtol, atol=atol)



"""
Functionality for grouping and filtering ScmRun objects
"""
import warnings
from collections.abc import Iterable

import numpy as np
from xarray.core import ops
from xarray.core.common import ImplementsArrayReduce


def _maybe_wrap_array(original, new_array):
    """
    Wrap a transformed array with ``__array_wrap__`` if it can be done safely.

    This lets us treat arbitrary functions that take and return ndarray objects
    like ufuncs, as long as they return an array with the same shape.
    """
    # in case func lost array's metadata
    if isinstance(new_array, np.ndarray) and new_array.shape == original.shape:
        return original.__array_wrap__(new_array)
    else:
        return new_array


class _GroupBy(ImplementsArrayReduce):
    def __init__(self, meta, groups, na_fill_value=-10000):
        m = meta.reset_index(drop=True)
        self.na_fill_value = float(na_fill_value)

        # Work around the bad handling of NaN values in groupbys
        if any([np.issubdtype(m[c].dtype, np.number) for c in m]):
            if (meta == na_fill_value).any(axis=None):
                raise ValueError(
                    "na_fill_value conflicts with data value. Choose a na_fill_value not in meta"
                )
            else:
                m = m.fillna(na_fill_value)

        self._grouper = m.groupby(list(groups), group_keys=True)

    def _iter_grouped(self):
        def _try_fill_value(v):
            try:
                if float(v) == float(self.na_fill_value):
                    return np.nan
            except ValueError:
                pass
            return v

        for indices in self._grouper.groups:
            if not isinstance(indices, Iterable) or isinstance(indices, str):
                indices = [indices]

            indices = [_try_fill_value(v) for v in indices]
            res = self.run.filter(**{k: v for k, v in zip(self.group_keys, indices)})
            if not len(res):
                raise ValueError(
                    "Empty group for {}".format(list(zip(self.group_keys, indices)))
                )
            yield res

    def __iter__(self):
        return self._iter_grouped()


class RunGroupBy(_GroupBy):
    """
    GroupBy object specialized to grouping ScmRun objects
    """

    def __init__(self, run, groups):
        self.run = run
        self.group_keys = groups
        super().__init__(run.meta, groups)

    def map(self, func, *args, **kwargs):
        """
        Apply a function to each group and append the results

        `func` is called like `func(ar, *args, **kwargs)` for each :obj:`ScmRun` ``ar``
        in this group. If the result of this function call is None, than it is
        excluded from the results.

        The results are appended together using :func:`run_append`. The function
        can change the size of the input :obj:`ScmRun` as long as :func:`run_append`
        can be applied to all results.

        Examples
        --------
        .. code:: python

            >>> def write_csv(arr):
            ...     variable = arr.get_unique_meta("variable")
            ...     arr.to_csv("out-{}.csv".format(variable)
            >>> df.groupby("variable").map(write_csv)

        Parameters
        ----------
        func : function
            Callable to apply to each timeseries.

        ``*args``
            Positional arguments passed to `func`.

        ``**kwargs``
            Used to call `func(ar, **kwargs)` for each array `ar`.

        Returns
        -------
        applied : :obj:`ScmRun`
            The result of splitting, applying and combining this array.
        """
        grouped = self._iter_grouped()
        applied = [
            _maybe_wrap_array(arr, func(arr, *args, **kwargs)) for arr in grouped
        ]
        return self._combine(applied)

    def _combine(self, applied):
        """
        Recombine the applied objects like the original.
        """
        from scmdata.run import run_append

        # Remove all None values
        applied = [df for df in applied if df is not None]

        if len(applied) == 0:
            return None
        else:
            return run_append(applied)

    def reduce(self, func, dim=None, axis=None, **kwargs):
        """
        Reduce the items in this group by applying `func` along some
        dimension(s).

        Parameters
        ----------
        func : function
            Function which can be called in the form
            `func(x, axis=axis, **kwargs)` to return the result of collapsing
            an np.ndarray over an integer valued axis.
        dim : `...`, str or sequence of str, optional
            Not used in this implementation
        axis : int or sequence of int, optional
            Axis(es) over which to apply `func`. Only one of the 'dimension'
            and 'axis' arguments can be supplied. If neither are supplied, then
            `func` is calculated over all dimension for each group item.
        **kwargs : dict
            Additional keyword arguments passed on to `func`.

        Returns
        -------
        reduced : :obj:`ScmRun`
            Array with summarized data and the indicated dimension(s)
            removed.
        """
        if dim is not None and dim != "time":
            raise ValueError("Only reduction along the time dimension is supported")

        def reduce_array(ar):
            return ar.reduce(func, dim, axis, **kwargs)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            return self.map(reduce_array)


ops.inject_reduce_methods(RunGroupBy)



"""
Database for handling large datasets in a performant, but flexible way

Data is chunked using unique combinations of metadata. This allows for the
database to expand as new data is added without having to change any of the
existing data.

Subsets of data are also able to be read without having to load all the data
and then filter. For example, one could save model results from a number of different
climate models and then load just the ``Surface Temperature`` data for all models.
"""
import glob
import os
import os.path
import pathlib
import shutil

import pandas as pd
import tqdm.autonotebook as tqdman

from scmdata import ScmRun, run_append


def ensure_dir_exists(fp):
    """
    Ensure directory exists

    Parameters
    ----------
    fp : str
        Filepath of which to ensure the directory exists
    """
    dir_to_check = os.path.dirname(fp)
    if not os.path.isdir(dir_to_check):
        try:
            os.makedirs(dir_to_check)
        except OSError:  # pragma: no cover
            # Prevent race conditions if multiple threads attempt to create dir at same time
            if not os.path.exists(dir_to_check):
                raise


def _check_is_subdir(root, d):
    root_path = pathlib.Path(root).resolve()
    out_path = pathlib.Path(d).resolve()

    is_subdir = root_path in out_path.parents
    # Sanity check that we never mangle anything outside of the root dir
    if not is_subdir:  # pragma: no cover
        raise AssertionError("{} not in {}".format(d, root))


class ScmDatabase:
    """
    On-disk database handler for outputs from SCMs

    Data is split into groups as specified by :attr:`levels`. This allows for fast
    reading and writing of new subsets of data when a single output file is no longer
    performant or data cannot all fit in memory.
    """

    def __init__(
        self, root_dir, levels=("climate_model", "variable", "region", "scenario"),
    ):
        """
        Initialise the database

        Parameters
        ----------
        root_dir : str
            The root directory of the database

        levels : tuple of str
            Specifies how the runs should be stored on disk.

            The data will be grouped by ``levels``. These levels should be adapted to
            best match the input data and desired access pattern. If there are any additional
            varying dimensions, they will be stored as dimensions.

        .. note::

            Creating a new :class:`ScmDatabase` does not modify any existing data on disk. To
            load an existing database ensure that the :attr:`root_dir` and :attr:`levels` are
            the same as the previous instance.
        """
        self._root_dir = root_dir
        self.levels = tuple(levels)

    def __repr__(self):
        return "<scmdata.database.SCMDatabase (root_dir: {}, levels: {})>".format(
            self._root_dir, self.levels
        )

    @property
    def root_dir(self):
        """
        Root directory of the database.

        Returns
        -------
        str
        """
        return self._root_dir

    @staticmethod
    def _get_disk_filename(inp):
        def safe_char(c):
            if c.isalnum() or c in "-/*_.":
                return c
            else:
                return "-"

        return "".join(safe_char(c) for c in inp)

    def save(self, scmrun):
        """
        Save data to the database

        The results are saved with one file for each unique combination of
        :attr:`levels` in a directory structure underneath ``root_dir``.

        Use :meth:`available_data` to see what data is available. Subsets of
        data can then be loaded as an :class:`scmdata.ScmRun` using :meth:`load`.

        Parameters
        ----------
        scmrun : :obj:`scmdata.ScmRun`
            Data to save.

            The timeseries in this run should have valid metadata for each
            of the columns specified in ``levels``.
        """
        for r in tqdman.tqdm(
            scmrun.groupby(self.levels), leave=False, desc="Saving to database",
        ):
            self._save_to_database_single_file(r)

    def _get_out_filepath(self, **levels):
        """
        Get filepath in which data has been saved

        The filepath is the root directory joined with the other information provided. The filepath
        is also cleaned to remove spaces and special characters.

        Parameters
        ----------
        levels: dict of str : str
            The unique value for each level in :attr:`levels'

        Returns
        -------
        str
            Path in which to save the data without spaces or special characters.

        Raises
        ------
        ValueError
            If no value is provided for level in :attr:`levels'
        """
        out_levels = []
        for level in self.levels:
            if level not in levels:
                raise ValueError("expected value for level: {}".format(level))
            out_levels.append(str(levels[level]))

        out_path = os.path.join(self._root_dir, *out_levels)
        out_fname = "__".join(out_levels) + ".nc"
        out_fname = os.path.join(out_path, out_fname)

        _check_is_subdir(self._root_dir, out_fname)

        return self._get_disk_filename(out_fname)

    def _save_to_database_single_file(self, scmrun):
        levels = {
            level: scmrun.get_unique_meta(level, no_duplicates=True).replace("/", "_")
            for level in self.levels
        }
        out_file = self._get_out_filepath(**levels)

        ensure_dir_exists(out_file)
        if os.path.exists(out_file):
            existing_run = ScmRun.from_nc(out_file)

            scmrun = run_append([existing_run, scmrun])

        # Check for required extra dimensions
        nunique_meta_vals = scmrun.meta.nunique()
        dimensions = nunique_meta_vals[nunique_meta_vals > 1].index.tolist()
        scmrun.to_nc(out_file, dimensions=dimensions)

    def load(self, **filters):
        """
        Load data from the database

        Parameters
        ----------
        filters: dict of str : str
            Filters for the data to load.

            Defaults to loading all values for a level if it isn't specified.

        Returns
        -------
        :class:`scmdata.ScmRun`
            Loaded data

        Raises
        ------
        ValueError
            If a filter for a level not in :attr:`levels` is specified

            If no data matching ``filters`` is found
        """
        for level in filters:
            if level not in self.levels:
                raise ValueError("Unknown level: {}".format(level))
            if "/" in filters[level]:
                filters[level] = filters[level].replace("/", "_")

        paths_to_load = [filters.get(level, "*") for level in self.levels]
        load_path = os.path.join(self._root_dir, *paths_to_load)
        glob_to_use = self._get_disk_filename(os.path.join(load_path, "*.nc"))
        load_files = glob.glob(glob_to_use, recursive=True)

        return run_append(
            [
                ScmRun.from_nc(f)
                for f in tqdman.tqdm(load_files, desc="Loading files", leave=False)
            ]
        )

    def delete(self, **filters):
        """
        Delete data from the database

        Parameters
        ----------
        filters: dict of str
            Filters for the data to load.

            Defaults to deleting all data if nothing is specified.

        Raises
        ------
        ValueError
            If a filter for a level not in :attr:`levels` is specified
        """
        for level in filters:
            if level not in self.levels:
                raise ValueError("Unknown level: {}".format(level))
            if "/" in filters[level]:
                filters[level] = filters[level].replace("/", "_")

        paths_to_load = [filters.get(level, "*") for level in self.levels]
        load_path = os.path.join(self._root_dir, *paths_to_load)
        glob_to_use = self._get_disk_filename(load_path)
        load_dirs = glob.glob(glob_to_use, recursive=True)

        for d in load_dirs:
            _check_is_subdir(self._root_dir, d)
            shutil.rmtree(d)

    def available_data(self):
        """
        Get all the data which is available to be loaded

        If metadata includes non-alphanumeric characters then it
        might appear modified in the returned table. The original
        metadata values can still be used to filter data.

        Returns
        -------
        :class:`pd.DataFrame`
        """
        load_path = os.path.join(self._root_dir, "**", "*.nc")
        all_files = glob.glob(load_path, recursive=True)

        file_meta = []
        for f in all_files:
            dirnames = f.split(os.sep)[:-1]
            file_meta.append(dirnames[-len(self.levels) :])

        data = pd.DataFrame(file_meta, columns=self.levels)

        return data.sort_values(by=data.columns.to_list()).reset_index(drop=True)



"""
ScmRun provides a high level analysis tool for simple climate model relevant
data. It provides a simple interface for reading/writing, subsetting and visualising
model data. ScmRuns are able to hold multiple model runs which aids in analysis of
ensembles of model runs.
"""
import copy
import datetime as dt
import functools
import numbers
import os
import warnings
from logging import getLogger
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
import numpy.testing as npt
import pandas as pd
import pint
from dateutil import parser
from openscm_units import unit_registry as ur
from xarray.core.ops import inject_binary_ops

from .errors import MissingRequiredColumnError, NonUniqueMetadataError
from .filters import (
    HIERARCHY_SEPARATOR,
    datetime_match,
    day_match,
    hour_match,
    month_match,
    pattern_match,
    years_match,
)
from .groupby import RunGroupBy
from .netcdf import inject_nc_methods
from .offsets import generate_range, to_offset
from .ops import inject_ops_methods
from .plotting import inject_plotting_methods
from .pyam_compat import IamDataFrame, LongDatetimeIamDataFrame
from .time import _TARGET_DTYPE, TimePoints, TimeseriesConverter
from .units import UnitConverter

_logger = getLogger(__name__)


MetadataType = Dict[str, Union[str, int, float]]


def _read_file(  # pylint: disable=missing-return-doc
    fnames: str, required_cols: Tuple[str], *args: Any, **kwargs: Any
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Prepare data to initialize :class:`ScmRun` from a file.

    Parameters
    ----------
    *args
        Passed to :func:`_read_pandas`.
    **kwargs
        Passed to :func:`_read_pandas`.

    Returns
    -------
    :obj:`pd.DataFrame`, :obj:`pd.DataFrame`
        First dataframe is the data. Second dataframe is metadata
    """
    _logger.info("Reading %s", fnames)

    return _format_data(_read_pandas(fnames, *args, **kwargs), required_cols)


def _read_pandas(
    fname: str, *args: Any, lowercase_cols=False, **kwargs: Any
) -> pd.DataFrame:
    """
    Read a file and return a :class:`pd.DataFrame`.

    Parameters
    ----------
    fname
        Path from which to read data

    lowercase_cols
        If True, convert the column names of the file to lowercase

    *args
        Passed to :func:`pd.read_excel` if :obj:`fname` ends with '.xls' or
        '.xslx, otherwise passed to :func:`pd.read_csv`.

    **kwargs
        Passed to :func:`pd.read_excel` if :obj:`fname` ends with '.xls' or
        '.xslx, otherwise passed to :func:`pd.read_csv`.

    Returns
    -------
    :obj:`pd.DataFrame`
        Read data

    Raises
    ------
    OSError
        Path specified by :obj:`fname` does not exist
    """
    if not os.path.exists(fname):
        raise OSError("no data file `{}` found!".format(fname))

    if fname.endswith("xlsx") or fname.endswith("xls"):
        _logger.debug("Assuming excel file")
        xl = pd.ExcelFile(fname)

        if len(xl.sheet_names) > 1 and "sheet_name" not in kwargs:
            kwargs["sheet_name"] = "data"

        df = pd.read_excel(fname, *args, **kwargs)

    else:
        _logger.debug("Reading with pandas read_csv")
        df = pd.read_csv(fname, *args, **kwargs)

    def _to_lower(c):
        if hasattr(c, "lower"):
            return c.lower()
        return c

    if lowercase_cols:
        df.columns = [_to_lower(c) for c in df.columns]

    return df


# pylint doesn't recognise return statements if they include ','
def _format_data(  # pylint: disable=missing-return-doc
    df: Union[pd.DataFrame, pd.Series], required_cols: Tuple[str]
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Prepare data to initialize :class:`ScmRun` from :class:`pd.DataFrame` or
    :class:`pd.Series`.

    See docstring of :func:`ScmRun.__init__` for details.

    Parameters
    ----------
    df
        Data to format.

    Returns
    -------
    :obj:`pd.DataFrame`, :obj:`pd.DataFrame`
        First dataframe is the data. Second dataframe is metadata.

    Raises
    ------
    ValueError
        Not all required metadata columns are present or the time axis cannot be
        understood
    """
    if isinstance(df, pd.Series):
        df = df.to_frame()

    # reset the index if meaningful entries are included there
    if list(df.index.names) != [None]:
        df.reset_index(inplace=True)

    if not set(required_cols).issubset(set(df.columns)):
        missing = list(set(required_cols) - set(df.columns))
        raise MissingRequiredColumnError(missing)

    # check whether data in wide or long format
    if "value" in df.columns:
        df, meta = _format_long_data(df, required_cols)
    else:
        df, meta = _format_wide_data(df, required_cols)

    # sort data
    df.sort_index(inplace=True)

    return df, meta


def _format_long_data(df, required_cols):
    # check if time column is given as `year` (int) or `time` (datetime)
    cols = set(df.columns)
    if "year" in cols and "time" not in cols:
        time_col = "year"
    elif "time" in cols and "year" not in cols:
        time_col = "time"
    else:
        msg = "invalid time format, must have either `year` or `time`!"
        raise ValueError(msg)

    required_cols = list(required_cols)
    extra_cols = list(set(cols) - set(required_cols + [time_col, "value"]))
    df = df.pivot_table(columns=required_cols + extra_cols, index=time_col).value
    meta = df.columns.to_frame(index=None)
    df.columns = meta.index

    return df, meta


def _format_wide_data(df, required_cols):
    cols = set(df.columns) - set(required_cols)
    time_cols, extra_cols = False, []
    for i in cols:
        # if in wide format, check if columns are years (int) or datetime
        if isinstance(i, dt.datetime):
            time_cols = True
        else:
            try:
                float(i)
                time_cols = True
            except (ValueError, TypeError):
                try:
                    try:
                        # most common format
                        dt.datetime.strptime(i, "%Y-%m-%d %H:%M:%S")
                    except ValueError:
                        # this is super slow so avoid if possible
                        parser.parse(str(i))  # if no ValueError, this is datetime
                    time_cols = True
                except ValueError:
                    extra_cols.append(i)  # some other string

    if not time_cols:
        msg = "invalid column format, must contain some time (int, float or datetime) columns!"
        raise ValueError(msg)

    all_cols = set(tuple(required_cols) + tuple(extra_cols))
    all_cols = list(all_cols)

    df_out = df.drop(all_cols, axis="columns").T
    df_out.index.name = "time"
    meta = df[all_cols].set_index(df_out.columns)

    return df_out, meta


def _from_ts(
    df: Any,
    required_cols: Tuple[str],
    index: Any = None,
    **columns: Union[str, bool, float, int, List],
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Prepare data to initialize :class:`ScmRun` from wide timeseries.

    See docstring of :func:`ScmRun.__init__` for details.

    Returns
    -------
    Tuple[pd.DataFrame, pd.DataFrame]
        First dataframe is the data. Second dataframe is metadata

    Raises
    ------
    ValueError
        Not all required columns are present
    """
    if not isinstance(df, pd.DataFrame):
        df = pd.DataFrame(df)
    if index is not None:
        if isinstance(index, np.ndarray):
            df.index = TimePoints(index).to_index()
        elif isinstance(index, TimePoints):
            df.index = index.to_index()
        else:
            df.index = index

    # format columns to lower-case and check that all required columns exist
    if not set(required_cols).issubset(columns.keys()):
        missing = list(set(required_cols) - set(columns.keys()))
        raise MissingRequiredColumnError(missing)

    df.index.name = "time"

    num_ts = len(df.columns)
    for c_name, col in columns.items():
        col_list = (
            [col] if isinstance(col, str) or not isinstance(col, Iterable) else col
        )

        if len(col_list) == num_ts:
            continue
        if len(col_list) != 1:
            error_msg = (
                "Length of column '{}' is incorrect. It should be length "
                "1 or {}".format(c_name, num_ts)
            )
            raise ValueError(error_msg)
        columns[c_name] = col_list * num_ts

    meta = pd.DataFrame(columns, index=df.columns)

    return df, meta


class BaseScmRun:  # pylint: disable=too-many-public-methods
    """
    Base class of a data container for timeseries data
    """

    required_cols = ("variable", "unit")
    """
    Required metadata columns

    This is the bare minimum columns which are expected. Attempting to create a run
    without the metadata columns specified by :attr:`required_cols` will raise a
    MissingRequiredColumnError
    """

    data_hierarchy_separator = HIERARCHY_SEPARATOR
    """
    str: String used to define different levels in our data hierarchies.

    By default we follow pyam and use "|". In such a case, emissions of |CO2| for
    energy from coal would be "Emissions|CO2|Energy|Coal".
    """

    def __init__(
        self,
        data: Any,
        index: Any = None,
        columns: Optional[Union[Dict[str, list], Dict[str, str]]] = None,
        metadata: Optional[MetadataType] = None,
        copy_data: bool = False,
        **kwargs: Any,
    ):
        """
        Initialize the container with timeseries data.

        Parameters
        ----------
        data: Union[ScmRun, IamDataFrame, pd.DataFrame, np.ndarray, str]
            If a :class:`ScmRun` object is provided, then a new
            :obj:`ScmRun` is created with a copy of the values and metadata from :obj:`data`.

            A :class:`pd.DataFrame` with IAMC-format data columns (the result
            from :func:`ScmRun.timeseries()`) can be provided without any additional
            :obj:`columns` and :obj:`index` information.

            If a numpy array of timeseries data is provided, :obj:`columns` and :obj:`index`
            must also be specified. The shape of the numpy array should be
            ``(n_times, n_series)`` where `n_times` is the number of timesteps and `n_series`
            is the number of time series.

            If a string is passed, data will be attempted to be read from file. Currently,
            reading from CSV, gzipped CSV and Excel formatted files is supported.

        index: np.ndarray
            If :obj:`index` is not ``None``, then the :obj:`index` is used as the timesteps
            for run. All timeseries in the run use the same set of timesteps.

            The values will be attempted to be converted to :class:`np.datetime[s]` values.
            Possible input formats include :

            * :obj:`datetime.datetime`
            * :obj:`int` Start of year
            * :obj:`float` Decimal year
            * :obj:`str` Uses :func:`dateutil.parser`. Slow and should be avoided if possible

            If :obj:`index` is ``None``, than the time index will be obtained from the
            :obj:`data` if possible.

        columns
            If None, ScmRun will attempt to infer the values from the source.
            Otherwise, use this dict to write the metadata for each timeseries in data.
            For each metadata key (e.g. "model", "scenario"), an array of values (one
            per time series) is expected. Alternatively, providing a list of length 1
            applies the same value to all timeseries in data. For example, if you had
            three timeseries from 'rcp26' for 3 different models 'model', 'model2' and
            'model3', the column dict would look like either 'col_1' or 'col_2':

            .. code:: python

                >>> col_1 = {
                    "scenario": ["rcp26"],
                    "model": ["model1", "model2", "model3"],
                    "region": ["unspecified"],
                    "variable": ["unspecified"],
                    "unit": ["unspecified"]
                }
                >>> col_2 = {
                    "scenario": ["rcp26", "rcp26", "rcp26"],
                    "model": ["model1", "model2", "model3"],
                    "region": ["unspecified"],
                    "variable": ["unspecified"],
                    "unit": ["unspecified"]
                }
                >>> assert pd.testing.assert_frame_equal(
                    ScmRun(d, columns=col_1).meta,
                    ScmRun(d, columns=col_2).meta
                )

        metadata:
            Optional dictionary of metadata for instance as a whole.

            This can be used to store information such as the longer-form information
            about a particular dataset, for example, dataset description or DOIs.

            Defaults to an empty :obj:`dict` if no default metadata are provided.

        copy_data: bool
            If True, an explicit copy of data is performed.

            .. note::
                The copy can be very expensive on large timeseries and should only be needed
                in cases where the original data is manipulated.

        **kwargs:
            Additional parameters passed to :func:`_read_file` to read files

        Raises
        ------
        ValueError
            * If you try to load from multiple files at once. If you wish to do this, please use :func:`scmdata.run.run_append` instead.
            * Not specifying :obj:`index` and :obj:`columns` if :obj:`data` is a :obj:`numpy.ndarray`

        :obj:`scmdata.errors.MissingRequiredColumn`
            If metadata for :attr:`required_cols` is not found

        TypeError
            Timeseries cannot be read from :obj:`data`
        """
        if isinstance(data, ScmRun):
            self._df = data._df.copy() if copy_data else data._df
            self._meta = data._meta
            self._time_points = TimePoints(data.time_points.values)
            if metadata is None:
                metadata = data.metadata.copy()
        else:
            if copy_data and hasattr(data, "copy"):
                data = data.copy()
            self._init_timeseries(data, index, columns, copy_data=copy_data, **kwargs)

        if self._duplicated_meta():
            raise NonUniqueMetadataError(self.meta)

        self.metadata = metadata.copy() if metadata is not None else {}

    def _init_timeseries(
        self,
        data,
        index: Any = None,
        columns: Optional[Dict[str, list]] = None,
        copy_data=False,
        **kwargs: Any,
    ):
        if isinstance(data, np.ndarray):
            if columns is None:
                raise ValueError("`columns` argument is required")
            if index is None:
                raise ValueError("`index` argument is required")

        if columns is not None:
            (_df, _meta) = _from_ts(
                data, index=index, required_cols=self.required_cols, **columns
            )
        elif isinstance(data, (pd.DataFrame, pd.Series)):
            (_df, _meta) = _format_data(data, self.required_cols)
        elif (IamDataFrame is not None) and isinstance(data, IamDataFrame):
            (_df, _meta) = _format_data(
                data.data.copy() if copy_data else data.data, self.required_cols
            )
        else:
            if not isinstance(data, str):
                if isinstance(data, list) and isinstance(data[0], str):
                    raise ValueError(
                        "Initialising from multiple files not supported, "
                        "use `scmdata.run.ScmRun.append()`"
                    )
                error_msg = "Cannot load {} from {}".format(type(self), type(data))
                raise TypeError(error_msg)

            (_df, _meta) = _read_file(data, required_cols=self.required_cols, **kwargs)

        self._time_points = TimePoints(_df.index.values)

        _df = _df.astype(float)
        self._df = _df
        self._df.index = self._time_points.to_index()
        self._meta = pd.MultiIndex.from_frame(_meta.astype("category"))

    def copy(self):
        """
        Return a :func:`copy.deepcopy` of self.

        Also creates copies the underlying Timeseries data

        Returns
        -------
        :obj:`ScmRun`
            :func:`copy.deepcopy` of ``self``
        """
        ret = copy.copy(self)
        ret._df = self._df.copy()
        ret._meta = self._meta.copy()
        ret.metadata = copy.copy(self.metadata)

        return ret

    def __len__(self) -> int:
        """
        Get the number of timeseries.
        """
        return self._df.shape[1]

    def __getitem__(self, key: Any) -> Any:
        """
        Get item of self with helpful direct access.

        Provides direct access to "time", "year" as well as the columns in :attr:`meta`.
        If key is anything else, the key will be applied to :attr:`_data`.
        """
        _key_check = (
            [key] if isinstance(key, str) or not isinstance(key, Iterable) else key
        )
        if key == "time":
            return pd.Series(self._time_points.to_index(), dtype="object")
        if key == "year":
            return pd.Series(self._time_points.years())
        if set(_key_check).issubset(self.meta_attributes):
            try:
                return self._meta_column(key).astype(
                    self._meta_column(key).cat.categories.dtype
                )
            except ValueError:
                return self._meta_column(key).astype(float)

        raise KeyError("[{}] is not in metadata".format(key))

    def __setitem__(
        self, key: str, value: Union[np.ndarray, list, int, float, str]
    ) -> Any:
        """
        Update metadata

        Notes
        -----
        If the meta values changes are applied to a filtered subset, the change will be reflected
        in the original :obj:`ScmRun` object.

        .. code:: python

            >>> df
            <scmdata.ScmRun (timeseries: 3, timepoints: 3)>
            Time:
                Start: 2005-01-01T00:00:00
                End: 2015-01-01T00:00:00
            Meta:
                   model     scenario region             variable   unit climate_model
                0  a_iam   a_scenario  World       Primary Energy  EJ/yr       a_model
                1  a_iam   a_scenario  World  Primary Energy|Coal  EJ/yr       a_model
                2  a_iam  a_scenario2  World       Primary Energy  EJ/yr       a_model
            >>> df["climate_model"] = ["a_model", "a_model", "b_model"]
            >>> df
            <scmdata.ScmRun (timeseries: 3, timepoints: 3)>
            Time:
                Start: 2005-01-01T00:00:00
                End: 2015-01-01T00:00:00
            Meta:
                   model     scenario region             variable   unit climate_model
                0  a_iam   a_scenario  World       Primary Energy  EJ/yr       a_model
                1  a_iam   a_scenario  World  Primary Energy|Coal  EJ/yr       a_model
                2  a_iam  a_scenario2  World       Primary Energy  EJ/yr       b_model
            >>> df2 = df.filter(variable="Primary Energy")
            >>> df2["pe_only"] = True
            >>> df2
            <scmdata.ScmRun (timeseries: 2, timepoints: 3)>
            Time:
                Start: 2005-01-01T00:00:00
                End: 2015-01-01T00:00:00
            Meta:
                   model     scenario region             variable   unit climate_model pe_only
                0  a_iam   a_scenario  World       Primary Energy  EJ/yr       a_model    True
                2  a_iam  a_scenario2  World       Primary Energy  EJ/yr       b_model    True
            >>> df
            <scmdata.ScmRun (timeseries: 3, timepoints: 3)>
            Time:
                Start: 2005-01-01T00:00:00
                End: 2015-01-01T00:00:00
            Meta:
                   model     scenario region             variable   unit climate_model pe_only
                0  a_iam   a_scenario  World       Primary Energy  EJ/yr       a_model    True
                1  a_iam   a_scenario  World  Primary Energy|Coal  EJ/yr       a_model     NaN
                2  a_iam  a_scenario2  World       Primary Energy  EJ/yr       b_model    True

        Parameters
        ----------
        key
            Column name

        value
            Values to write

            If a list of values is provided, then the length of that :obj:`value` must be the same as the number of timeseries

        Raises
        ------
        ValueError
            If the length of :obj:`meta` is inconsistent with the number of timeseries
        """
        meta = np.atleast_1d(value)
        if key == "time":
            self._time_points = TimePoints(meta)
            self._df.index = self._time_points.to_index()
        else:
            if len(meta) == 1:
                new_meta = self._meta.to_frame()
                new_meta[key] = meta[0]
                self._meta = pd.MultiIndex.from_frame(new_meta.astype("category"))
            elif len(meta) == len(self):
                new_meta_index = self._meta.to_frame(index=False)
                new_meta_index[key] = pd.Series(meta, dtype="category")
                self._meta = pd.MultiIndex.from_frame(new_meta_index)
            else:
                raise ValueError(
                    "Invalid length for metadata, `{}`, must be 1 or equal to the number of timeseries, `{}`".format(
                        len(meta), len(self)
                    )
                )

        if self._duplicated_meta():
            raise NonUniqueMetadataError(self.meta)

    def __repr__(self):
        def _indent(s):
            lines = ["\t" + line for line in s.split("\n")]
            return "\n".join(lines)

        meta_str = _indent(self.meta.__repr__())
        time_str = [
            "Start: {}".format(self.time_points.values[0]),
            "End: {}".format(self.time_points.values[-1]),
        ]
        time_str = _indent("\n".join(time_str))
        return "<scmdata.ScmRun (timeseries: {}, timepoints: {})>\nTime:\n{}\nMeta:\n{}".format(
            len(self), len(self.time_points), time_str, meta_str
        )

    @staticmethod
    def _binary_op(
        f: Callable[..., Any], reflexive=False, **kwargs,
    ) -> Callable[..., "ScmRun"]:
        @functools.wraps(f)
        def func(self, other):
            if isinstance(other, ScmRun):
                return NotImplemented

            is_scalar = isinstance(other, (numbers.Number, pint.Quantity))
            if not is_scalar:
                other_ndim = len(other.shape)
                if other_ndim == 1:
                    if other.shape[0] != self.shape[1]:
                        raise ValueError(
                            "only vectors with the same number of timesteps "
                            "as self ({}) are supported".format(self.shape[1])
                        )
                else:
                    raise ValueError(
                        "operations with {}d data are not supported".format(other_ndim)
                    )

            def _perform_op(df):
                if isinstance(other, pint.Quantity):
                    try:
                        data = df.values * ur(df.get_unique_meta("unit", True))
                        use_pint = True
                    except KeyError:
                        # let Pint assume dimensionless and raise an error as
                        # necessary
                        data = df.values
                        use_pint = False
                else:
                    data = df.values
                    use_pint = False

                res = []
                for v in data:
                    if not reflexive:
                        res.append(f(v, other))
                    else:
                        res.append(f(other, v))
                res = np.vstack(res)

                if use_pint:
                    df._df.values[:] = res.magnitude.T
                    df["unit"] = str(res.units)
                else:
                    df._df.values[:] = res.T
                return df

            return self.copy().groupby("unit").map(_perform_op)

        return func

    def drop_meta(self, columns: Union[list, str], inplace: Optional[bool] = False):
        """
        Drop meta columns out of the Run

        Parameters
        ----------
        columns
            The column or columns to drop
        inplace
            If True, do operation inplace and return None.

        Raises
        ------
        KeyError
            If any of the columns do not exist in the meta :class:`DataFrame`
        """
        if inplace:
            ret = self
        else:
            ret = self.copy()

        if isinstance(columns, str):
            columns = [columns]

        existing_cols = ret.meta_attributes
        for c in columns:
            if c not in existing_cols:
                raise KeyError(c)
            if c in self.required_cols:
                raise MissingRequiredColumnError([c])
        for c in columns:
            ret._meta = ret._meta.droplevel(c)

        if ret._duplicated_meta():
            raise NonUniqueMetadataError(ret.meta)

        if not inplace:
            return ret

    @property
    def meta_attributes(self):
        """
        Get a list of all meta keys

        Returns
        -------
        list
            Sorted list of meta keys
        """
        return sorted(list(self._meta.names))

    @property
    def time_points(self):
        """
        Time points of the data

        Returns
        -------
        :obj:`scmdata.time.TimePoints`
        """
        return self._time_points

    def timeseries(
        self, meta=None, check_duplicated=True, time_axis=None, drop_all_nan_times=False
    ):
        """
        Return the data with metadata as a :obj:`pd.DataFrame`.

        Parameters
        ----------
        meta : list[str]
            The list of meta columns that will be included in the output's
            MultiIndex. If None (default), then all metadata will be used.

        check_duplicated : bool
            If True, an exception is raised if any of the timeseries have
            duplicated metadata

        time_axis : {None, "year", "year-month", "days since 1970-01-01", "seconds since 1970-01-01"}
            Time axis to use for the output's columns. If ``None``,
            :class:`datetime.datetime` objects will be used. If ``"year"``, the
            year of each time point  will be used. If ``"year-month"``, the year
            plus (month - 0.5) / 12  will be used. If
            ``"days since 1970-01-01"``, the number of days since 1st Jan 1970
            will be used (calculated using the :mod:`datetime` module). If
            ``"seconds since 1970-01-01"``, the number of seconds  since 1st Jan
            1970 will be used (calculated using the :mod:`datetime` module).

        drop_all_nan_times : bool
            Should time points which contain only nan values be dropped? This operation is applied
            after any transforms introduced by the value of ``time_axis``.

        Returns
        -------
        :obj:`pd.DataFrame`
            DataFrame with datetimes as columns and timeseries as rows.
            Metadata is in the index.

        Raises
        ------
        :class:`NonUniqueMetadataError`
            If the metadata are not unique between timeseries and
            ``check_duplicated`` is ``True``

        NotImplementedError
            The value of `time_axis` is not recognised

        ValueError
            The value of `time_axis` would result in columns which aren't unique
        """
        df = self._df.T
        _meta = self.meta if meta is None else self.meta[meta]

        if check_duplicated and self._duplicated_meta(meta=_meta):
            raise NonUniqueMetadataError(_meta)

        if time_axis is None:
            columns = self._time_points.to_index()
        elif time_axis == "year":
            columns = self._time_points.years()
        elif time_axis == "year-month":
            columns = (
                self._time_points.years() + (self._time_points.months() - 0.5) / 12
            )
        elif time_axis == "days since 1970-01-01":

            def calc_days(x):
                ref = np.array(["1970-01-01"], dtype=_TARGET_DTYPE)[0]

                return (x - ref).astype("timedelta64[D]")

            columns = calc_days(self._time_points.values).astype(int)

        elif time_axis == "seconds since 1970-01-01":

            def calc_seconds(x):
                ref = np.array(["1970-01-01"], dtype=_TARGET_DTYPE)[0]

                return x - ref

            columns = calc_seconds(self._time_points.values).astype(int)

        else:
            raise NotImplementedError("time_axis = '{}'".format(time_axis))

        if len(np.unique(columns)) != len(columns):
            raise ValueError(
                "Ambiguous time values with time_axis = '{}'".format(time_axis)
            )

        df.columns = pd.Index(columns, name="time")
        df.index = pd.MultiIndex.from_frame(_meta)

        if drop_all_nan_times:
            df = df.dropna(how="all", axis="columns")

        return df

    def _duplicated_meta(self, meta=None):
        _meta = self._meta if meta is None else meta

        return _meta.duplicated().any()

    def long_data(self, time_axis=None):
        """
        Return data in long form, particularly useful for plotting with seaborn

        Parameters
        ----------
        time_axis : {None, "year", "year-month", "days since 1970-01-01", "seconds since 1970-01-01"}
            Time axis to use for the output's columns. If `None`,
            :class:`datetime.datetime` objects will be used. If `"year"`, the
            year of each time point  will be used. If `"year-month", the year
            plus (month - 0.5) / 12  will be used. If
            `"days since 1970-01-01"`, the number of days  since 1st Jan 1970
            will be used (calculated using the ``datetime``  module). If
            `"seconds since 1970-01-01"`, the number of seconds  since 1st Jan
            1970 will be used (calculated using the ``datetime`` module).

        Returns
        -------
        :obj:`pd.DataFrame`
            :obj:`pd.DataFrame` containing the data in 'long form' (i.e. one observation per row).
        """
        out = self.timeseries(time_axis=time_axis).stack()
        out.name = "value"
        out = out.to_frame().reset_index()

        return out

    @property
    def shape(self) -> tuple:
        """
        Get the shape of the underlying data as ``(num_timeseries, num_timesteps)``

        Returns
        -------
        tuple of int
        """
        return self._df.T.shape

    @property
    def values(self) -> np.ndarray:
        """
        Timeseries values without metadata

        The values are returned such that each row is a different
        timeseries being a row and each column is a different time (although
        no time information is included as a plain :obj:`np.ndarray` is
        returned).

        Returns
        -------
        np.ndarray
            The array in the same shape as :py:obj:`ScmRun.shape`, that is
            ``(num_timeseries, num_timesteps)``.
        """
        return self._df.values.T

    @property
    def empty(self) -> bool:
        """
        Indicate whether :obj:`ScmRun` is empty i.e. contains no data

        Returns
        -------
        bool
            If :obj:`ScmRun` is empty, return ``True``, if not return ``False``
        """
        return np.equal(len(self), 0)

    @property
    def meta(self) -> pd.DataFrame:
        """
        Metadata
        """
        df = pd.DataFrame(
            self._meta.to_list(), columns=self._meta.names, index=self._df.columns
        )

        return df[sorted(df.columns)]

    def _meta_column(self, col) -> pd.Series:
        out = self._meta.get_level_values(col)
        return pd.Series(out, name=col, index=self._df.columns)

    def filter(
        self,
        keep: bool = True,
        inplace: bool = False,
        log_if_empty: bool = True,
        **kwargs: Any,
    ):
        """
        Return a filtered ScmRun (i.e., a subset of the data).

        .. code:: python

            >>> df
            <scmdata.ScmRun (timeseries: 3, timepoints: 3)>
            Time:
                Start: 2005-01-01T00:00:00
                End: 2015-01-01T00:00:00
            Meta:
                   model     scenario region             variable   unit climate_model
                0  a_iam   a_scenario  World       Primary Energy  EJ/yr       a_model
                1  a_iam   a_scenario  World  Primary Energy|Coal  EJ/yr       a_model
                2  a_iam  a_scenario2  World       Primary Energy  EJ/yr       a_model
                [3 rows x 7 columns]

            >>> df.filter(scenario="a_scenario")
            <scmdata.ScmRun (timeseries: 2, timepoints: 3)>
            Time:
                Start: 2005-01-01T00:00:00
                End: 2015-01-01T00:00:00
            Meta:
                   model     scenario region             variable   unit climate_model
                0  a_iam   a_scenario  World       Primary Energy  EJ/yr       a_model
                1  a_iam   a_scenario  World  Primary Energy|Coal  EJ/yr       a_model
                [2 rows x 7 columns]

            >>> df.filter(scenario="a_scenario", keep=False)
            <scmdata.ScmRun (timeseries: 1, timepoints: 3)>
            Time:
                Start: 2005-01-01T00:00:00
                End: 2015-01-01T00:00:00
            Meta:
                   model     scenario region             variable   unit climate_model
                2  a_iam  a_scenario2  World       Primary Energy  EJ/yr       a_model
                [1 rows x 7 columns]

            >>> df.filter(level=1)
            <scmdata.ScmRun (timeseries: 2, timepoints: 3)>
            Time:
                Start: 2005-01-01T00:00:00
                End: 2015-01-01T00:00:00
            Meta:
                   model     scenario region             variable   unit climate_model
                0  a_iam   a_scenario  World       Primary Energy  EJ/yr       a_model
                2  a_iam  a_scenario2  World       Primary Energy  EJ/yr       a_model
                [2 rows x 7 columns]

            >>> df.filter(year=range(2000, 2011))
            <scmdata.ScmRun (timeseries: 3, timepoints: 2)>
            Time:
                Start: 2005-01-01T00:00:00
                End: 2010-01-01T00:00:00
            Meta:
                   model     scenario region             variable   unit climate_model
                0  a_iam   a_scenario  World       Primary Energy  EJ/yr       a_model
                1  a_iam   a_scenario  World  Primary Energy|Coal  EJ/yr       a_model
                2  a_iam  a_scenario2  World       Primary Energy  EJ/yr       a_model
                [2 rows x 7 columns]

        Parameters
        ----------
        keep
            If True, keep all timeseries satisfying the filters, otherwise drop all the
            timeseries satisfying the filters

        inplace
            If True, do operation inplace and return None

        log_if_empty
            If ``True``, log a warning level message if the result is empty.

        **kwargs
            Argument names are keys with which to filter, values are used to do the
            filtering. Filtering can be done on:

            - all metadata columns with strings, "*" can be used as a wildcard in search
              strings

            - 'level': the maximum "depth" of IAM variables (number of hierarchy levels,
              excluding the strings given in the 'variable' argument)

            - 'time': takes a :class:`datetime.datetime` or list of
              :class:`datetime.datetime`'s
              TODO: default to np.datetime64

            - 'year', 'month', 'day', hour': takes an :class:`int` or list of
              :class:`int`'s ('month' and 'day' also accept :class:`str` or list of
              :class:`str`)

            If ``regexp=True`` is included in :obj:`kwargs` then the pseudo-regexp
            syntax in :func:`pattern_match` is disabled.

        Returns
        -------
        :obj:`ScmRun`
            If not ``inplace``, return a new instance with the filtered data.
        """
        ret = copy.copy(self) if not inplace else self

        if len(ret):
            _keep_times, _keep_rows = self._apply_filters(kwargs)
            if not keep and sum(~_keep_rows) and sum(~_keep_times):
                raise ValueError(
                    "If keep==False, filtering cannot be performed on the temporal axis "
                    "and with metadata at the same time"
                )

            reduce_times = (~_keep_times).sum() > 0
            reduce_rows = (~_keep_rows).sum() > 0

            if not keep:
                if reduce_times:
                    _keep_times = ~_keep_times
                if reduce_rows:
                    _keep_rows = ~_keep_rows
                if not reduce_rows and not reduce_times:
                    _keep_times = _keep_times * False
                    _keep_rows = _keep_rows * False

            ret._df = ret._df.loc[_keep_times, _keep_rows]
            ret._meta = ret._meta[_keep_rows]
            ret["time"] = self.time_points.values[_keep_times]

        if log_if_empty and ret.empty:
            _logger.warning("Filtered ScmRun is empty!", stack_info=True)

        if not inplace:
            return ret

        return None

    # pylint doesn't recognise ',' in returns type definition
    def _apply_filters(  # pylint: disable=missing-return-doc
        self, filters: Dict
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Determine rows to keep in data for given set of filters.

        Parameters
        ----------
        filters
            Dictionary of filters ``({col: values}})``; uses a pseudo-regexp syntax by
            default but if ``filters["regexp"]`` is ``True``, regexp is used directly.

        Returns
        -------
        :obj:`np.ndarray` of :class:`bool`, :obj:`np.ndarray` of :class:`bool`
            Two boolean :class:`np.ndarray`'s. The first contains the columns to keep
            (i.e. which time points to keep). The second contains the rows to keep (i.e.
            which metadata matched the filters).

        Raises
        ------
        ValueError
            Filtering cannot be performed on requested column
        """
        regexp = filters.pop("regexp", False)
        keep_ts = np.array([True] * len(self.time_points))
        keep_meta = np.array([True] * len(self))

        # filter by columns and list of values
        for col, values in filters.items():
            if col in self._meta.names:
                if col == "variable":
                    level = filters["level"] if "level" in filters else None
                else:
                    level = None

                keep_meta &= pattern_match(
                    self._meta.get_level_values(col),
                    values,
                    level=level,
                    regexp=regexp,
                    separator=self.data_hierarchy_separator,
                )

            elif col == "level":
                if "variable" not in filters.keys():
                    keep_meta &= pattern_match(
                        self._meta.get_level_values("variable"),
                        "*",
                        level=values,
                        regexp=regexp,
                        separator=self.data_hierarchy_separator,
                    )
                # else do nothing as level handled in variable filtering

            elif col == "year":
                keep_ts &= years_match(self._time_points.years(), values)

            elif col == "month":
                keep_ts &= month_match(self._time_points.months(), values)

            elif col == "day":
                keep_ts &= self._day_match(values)

            elif col == "hour":
                keep_ts &= hour_match(self._time_points.hours(), values)

            elif col == "time":
                keep_ts &= datetime_match(self._time_points.values, values)

            else:
                raise ValueError("filter by `{}` not supported".format(col))

        return keep_ts, keep_meta

    def _day_match(self, values):
        if isinstance(values, str):
            wday = True
        elif isinstance(values, list) and isinstance(values[0], str):
            wday = True
        else:
            wday = False

        if wday:
            days = self._time_points.weekdays()
        else:  # ints or list of ints
            days = self._time_points.days()

        return day_match(days, values)

    def head(self, *args, **kwargs):
        """
        Return head of :func:`self.timeseries()`.

        Parameters
        ----------
        *args
            Passed to :func:`self.timeseries().head()`

        **kwargs
            Passed to :func:`self.timeseries().head()`

        Returns
        -------
        :obj:`pd.DataFrame`
            Tail of :func:`self.timeseries()`
        """
        return self.timeseries().head(*args, **kwargs)

    def tail(self, *args: Any, **kwargs: Any) -> pd.DataFrame:
        """
        Return tail of :func:`self.timeseries()`.

        Parameters
        ----------
        *args
            Passed to :func:`self.timeseries().tail()`

        **kwargs
            Passed to :func:`self.timeseries().tail()`

        Returns
        -------
        :obj:`pd.DataFrame`
            Tail of :func:`self.timeseries()`
        """
        return self.timeseries().tail(*args, **kwargs)

    def get_unique_meta(
        self, meta: str, no_duplicates: Optional[bool] = False,
    ) -> Union[List[Any], Any]:
        """
        Get unique values in a metadata column.

        Parameters
        ----------
        meta
            Column to retrieve metadata for

        no_duplicates
            Should I raise an error if there is more than one unique value in the
            metadata column?

        Raises
        ------
        ValueError
            There is more than one unique value in the metadata column and
            ``no_duplicates`` is ``True``.

        Returns
        -------
        [List[Any], Any]
            List of unique metadata values. If ``no_duplicates`` is ``True`` the
            metadata value will be returned (rather than a list).
        """
        vals = self._meta.get_level_values(meta).unique().to_list()
        if no_duplicates:
            if len(vals) != 1:
                raise ValueError(
                    "`{}` column is not unique (found values: {})".format(meta, vals)
                )

            return vals[0]

        return vals

    def interpolate(
        self,
        target_times: Union[np.ndarray, List[Union[dt.datetime, int]]],
        interpolation_type: str = "linear",
        extrapolation_type: str = "linear",
    ):
        """
        Interpolate the dataframe onto a new time frame.

        Parameters
        ----------
        target_times
            Time grid onto which to interpolate
        interpolation_type: str
            Interpolation type. Options are 'linear'
        extrapolation_type: str or None
            Extrapolation type. Options are None, 'linear' or 'constant'

        Returns
        -------
        :obj:`ScmRun`
            A new :class:`ScmRun` containing the data interpolated onto the
            :obj:`target_times` grid
        """
        # pylint: disable=protected-access

        target_times = np.asarray(target_times, dtype="datetime64[s]")

        res = self.copy()

        target_times = TimePoints(target_times)

        timeseries_converter = TimeseriesConverter(
            self.time_points.values,
            target_times.values,
            interpolation_type=interpolation_type,
            extrapolation_type=extrapolation_type,
        )
        target_data = np.zeros((len(target_times), len(res)))

        # TODO: Extend TimeseriesConverter to handle 2d inputs
        for i in range(len(res)):
            target_data[:, i] = timeseries_converter.convert_from(
                res._df.iloc[:, i].values
            )
        res._df = pd.DataFrame(
            target_data, columns=res._df.columns, index=target_times.to_index()
        )
        res._time_points = target_times

        return res

    def resample(self, rule: str = "AS", **kwargs: Any):
        """
        Resample the time index of the timeseries data onto a custom grid.

        This helper function allows for values to be easily interpolated onto annual or
        monthly timesteps using the rules='AS' or 'MS' respectively. Internally, the
        interpolate function performs the regridding.

        Parameters
        ----------
        rule
            See the pandas `user guide
            <http://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects>`_
            for a list of options. Note that Business-related offsets such as
            "BusinessDay" are not supported.

        **kwargs
            Other arguments to pass through to :func:`interpolate`

        Returns
        -------
        :obj:`ScmRun`
            New :class:`ScmRun` instance on a new time index

        Examples
        --------
        Resample a dataframe to annual values

        >>> scm_df = ScmRun(
        ...     pd.Series([1, 2, 10], index=(2000, 2001, 2009)),
        ...     columns={
        ...         "model": ["a_iam"],
        ...         "scenario": ["a_scenario"],
        ...         "region": ["World"],
        ...         "variable": ["Primary Energy"],
        ...         "unit": ["EJ/y"],
        ...     }
        ... )
        >>> scm_df.timeseries().T
        model             a_iam
        scenario     a_scenario
        region            World
        variable Primary Energy
        unit               EJ/y
        year
        2000                  1
        2010                 10

        An annual timeseries can be the created by interpolating to the start of years
        using the rule 'AS'.

        >>> res = scm_df.resample('AS')
        >>> res.timeseries().T
        model                        a_iam
        scenario                a_scenario
        region                       World
        variable            Primary Energy
        unit                          EJ/y
        time
        2000-01-01 00:00:00       1.000000
        2001-01-01 00:00:00       2.001825
        2002-01-01 00:00:00       3.000912
        2003-01-01 00:00:00       4.000000
        2004-01-01 00:00:00       4.999088
        2005-01-01 00:00:00       6.000912
        2006-01-01 00:00:00       7.000000
        2007-01-01 00:00:00       7.999088
        2008-01-01 00:00:00       8.998175
        2009-01-01 00:00:00      10.00000

        >>> m_df = scm_df.resample('MS')
        >>> m_df.timeseries().T
        model                        a_iam
        scenario                a_scenario
        region                       World
        variable            Primary Energy
        unit                          EJ/y
        time
        2000-01-01 00:00:00       1.000000
        2000-02-01 00:00:00       1.084854
        2000-03-01 00:00:00       1.164234
        2000-04-01 00:00:00       1.249088
        2000-05-01 00:00:00       1.331204
        2000-06-01 00:00:00       1.416058
        2000-07-01 00:00:00       1.498175
        2000-08-01 00:00:00       1.583029
        2000-09-01 00:00:00       1.667883
                                    ...
        2008-05-01 00:00:00       9.329380
        2008-06-01 00:00:00       9.414234
        2008-07-01 00:00:00       9.496350
        2008-08-01 00:00:00       9.581204
        2008-09-01 00:00:00       9.666058
        2008-10-01 00:00:00       9.748175
        2008-11-01 00:00:00       9.833029
        2008-12-01 00:00:00       9.915146
        2009-01-01 00:00:00      10.000000
        [109 rows x 1 columns]


        Note that the values do not fall exactly on integer values as not all years are
        exactly the same length.

        References
        ----------
        See the pandas documentation for
        `resample <http://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.resample.html>`
        for more information about possible arguments.
        """
        orig_dts = self["time"]
        target_dts = generate_range(
            orig_dts.iloc[0], orig_dts.iloc[-1], to_offset(rule)
        )
        return self.interpolate(list(target_dts), **kwargs)

    def time_mean(self, rule: str):
        """
        Take time mean of self

        Note that this method will not copy the ``metadata`` attribute to the returned
        value.

        Parameters
        ----------
        rule : ["AC", "AS", "A"]
            How to take the time mean. The names reflect the pandas
            `user guide <http://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects>`_
            where they can, but only the options
            given above are supported. For clarity, if ``rule`` is ``'AC'``, then the
            mean is an annual mean i.e. each time point in the result is the mean of
            all values for that particular year. If ``rule`` is ``'AS'``, then the
            mean is an annual mean centred on the beginning of the year i.e. each time
            point in the result is the mean of all values from July 1st in the
            previous year to June 30 in the given year. If ``rule`` is ``'A'``, then
            the mean is an annual mean centred on the end of the year i.e. each time
            point in the result is the mean of all values from July 1st of the given
            year to June 30 in the next year.

        Returns
        -------
        :obj:`ScmRun`
            The time mean of ``self``.
        """
        if rule == "AS":

            def group_annual_mean_beginning_of_year(x):
                if x.month <= 6:
                    return x.year
                return x.year + 1

            ts_resampled = (
                self.timeseries()
                .T.groupby(group_annual_mean_beginning_of_year)
                .mean()
                .T
            )
            ts_resampled.columns = ts_resampled.columns.map(
                lambda x: dt.datetime(x, 1, 1)
            )
            return type(self)(ts_resampled)

        if rule == "AC":

            def group_annual_mean(x):
                return x.year

            ts_resampled = self.timeseries().T.groupby(group_annual_mean).mean().T
            ts_resampled.columns = ts_resampled.columns.map(
                lambda x: dt.datetime(x, 7, 1)
            )
            return type(self)(ts_resampled)

        if rule == "A":

            def group_annual_mean_end_of_year(x):
                if x.month >= 7:
                    return x.year
                return x.year - 1

            ts_resampled = (
                self.timeseries().T.groupby(group_annual_mean_end_of_year).mean().T
            )
            ts_resampled.columns = ts_resampled.columns.map(
                lambda x: dt.datetime(x, 12, 31)
            )
            return type(self)(ts_resampled)

        raise ValueError("`rule` = `{}` is not supported".format(rule))

    def process_over(
        self,
        cols: Union[str, List[str]],
        operation: str,
        na_override=-1e6,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """
        Process the data over the input columns.

        Parameters
        ----------
        cols
            Columns to perform the operation on. The timeseries will be grouped by all
            other columns in :attr:`meta`.

        operation : ['median', 'mean', 'quantile']
            The operation to perform. This uses the equivalent pandas function. Note
            that quantile means the value of the data at a given point in the cumulative
            distribution of values at each point in the timeseries, for each timeseries
            once the groupby is applied. As a result, using ``q=0.5`` is is the same as
            taking the median and not the same as taking the mean/average.

        na_override: [int, float]
            Convert any nan value in the timeseries meta to this value during processsing.
            The meta values converted back to nan's before the dataframe is returned. This
            should not need to be changed unless the existing metadata clashes with the
            default na_override value.

            This functionality is disabled if na_override is None, but may result incorrect
            results if the timeseries meta includes any nan's.

        **kwargs
            Keyword arguments to pass to the pandas operation

        Returns
        -------
        :obj:`pd.DataFrame`
            The quantiles of the timeseries, grouped by all columns in :attr:`meta`
            other than :obj:`cols`

        Raises
        ------
        ValueError
            If the operation is not one of ['median', 'mean', 'quantile']

            If the value of na_override clashes with any existing metadata
        """
        cols = [cols] if isinstance(cols, str) else cols
        ts = self.timeseries()
        if na_override is not None:
            ts_idx = ts.index.to_frame()
            if ts_idx[ts_idx == na_override].any().any():
                raise ValueError(
                    "na_override clashes with existing meta: {}".format(na_override)
                )
            ts.index = pd.MultiIndex.from_frame(ts_idx.fillna(na_override))
        group_cols = list(set(ts.index.names) - set(cols))
        grouper = ts.groupby(group_cols)

        if operation == "median":
            res = grouper.median(**kwargs)
        elif operation == "mean":
            res = grouper.mean(**kwargs)
        elif operation == "quantile":
            res = grouper.quantile(**kwargs)
        else:
            raise ValueError("operation must be one of ['median', 'mean', 'quantile']")

        if na_override is not None:
            idx_df = res.index.to_frame()
            idx_df[idx_df == na_override] = np.nan
            res.index = pd.MultiIndex.from_frame(idx_df)
        return res

    def groupby(self, *group):
        """
        Group the object by unique metadata

        Enables iteration over groups of data. For example, to iterate over each scenario in the object

        .. code:: python

            >>> for group in df.groupby("scenario"):
            >>>    print(group)
            <scmdata.ScmRun (timeseries: 2, timepoints: 3)>
            Time:
                Start: 2005-01-01T00:00:00
                End: 2015-01-01T00:00:00
            Meta:
                   model    scenario region             variable   unit climate_model
                0  a_iam  a_scenario  World       Primary Energy  EJ/yr       a_model
                1  a_iam  a_scenario  World  Primary Energy|Coal  EJ/yr       a_model
            <scmdata.ScmRun (timeseries: 1, timepoints: 3)>
            Time:
                Start: 2005-01-01T00:00:00
                End: 2015-01-01T00:00:00
            Meta:
                   model     scenario region        variable   unit climate_model
                2  a_iam  a_scenario2  World  Primary Energy  EJ/yr       a_model

        Parameters
        ----------
        group: str or list of str
            Columns to group by

        Returns
        -------
        :obj:`RunGroupBy`
            See the documentation for :class:`RunGroupBy` for more information

        """
        if len(group) == 1 and not isinstance(group[0], str):
            group = tuple(group[0])
        return RunGroupBy(self, group)

    def convert_unit(
        self,
        unit: str,
        context: Optional[str] = None,
        inplace: bool = False,
        **kwargs: Any,
    ):
        """
        Convert the units of a selection of timeseries.

        Uses :class:`scmdata.units.UnitConverter` to perform the conversion.

        Parameters
        ----------
        unit
            Unit to convert to. This must be recognised by
            :class:`~openscm.units.UnitConverter`.

        context
            Context to use for the conversion i.e. which metric to apply when performing
            CO2-equivalent calculations. If ``None``, no metric will be applied and
            CO2-equivalent calculations will raise :class:`DimensionalityError`.

        inplace
            If True, apply the conversion inplace and return None

        **kwargs
            Extra arguments which are passed to :func:`~ScmRun.filter` to
            limit the timeseries which are attempted to be converted. Defaults to
            selecting the entire ScmRun, which will likely fail.

        Returns
        -------
        :obj:`ScmRun`
            If :obj:`inplace` is not ``False``, a new :class:`ScmRun` instance
            with the converted units.

        Notes
        -----
        If ``context`` is not ``None``, then the context used for the conversion will
        be checked against any existing metadata and, if the conversion is valid,
        stored in the output's metadata.

        Raises
        ------
        ValueError
            ``"unit_context"`` is already included in ``self``'s :meth:`meta_attributes`
            and it does not match ``context`` for the variables to be converted.
        """
        # pylint: disable=protected-access
        if inplace:
            ret = self
        else:
            ret = self.copy()

        to_convert_filtered = ret.filter(**kwargs, log_if_empty=False)
        to_not_convert_filtered = ret.filter(**kwargs, keep=False, log_if_empty=False)

        already_correct_unit = to_convert_filtered.filter(unit=unit, log_if_empty=False)
        if (
            "unit_context" in already_correct_unit.meta_attributes
            and not already_correct_unit.empty
        ):
            self._check_unit_context(already_correct_unit, context)

        to_convert = to_convert_filtered.filter(
            unit=unit, log_if_empty=False, keep=False
        )
        to_not_convert = run_append([to_not_convert_filtered, already_correct_unit,])

        if "unit_context" in to_convert.meta_attributes and not to_convert.empty:
            self._check_unit_context(to_convert, context)

        if context is not None:
            to_convert["unit_context"] = context

        if "unit_context" not in to_not_convert.meta_attributes and context is not None:
            to_not_convert["unit_context"] = None

        def apply_units(group):
            orig_unit = group.get_unique_meta("unit", no_duplicates=True)
            uc = UnitConverter(orig_unit, unit, context=context)

            group._df.values[:] = uc.convert_from(group._df.values)
            group["unit"] = unit

            return group

        ret = to_convert
        if not to_convert.empty:
            ret = ret.groupby("unit").map(apply_units)

        ret = run_append([ret, to_not_convert], inplace=inplace)
        if not inplace:
            return ret

    @staticmethod
    def _check_unit_context(dat, context):
        unit_context = dat.get_unique_meta("unit_context")

        # check if contexts don't match, unless the context is nan
        non_matching_contexts = len(unit_context) > 1 or unit_context[0] != context
        if isinstance(unit_context[0], float):
            non_matching_contexts &= not np.isnan(unit_context[0])

        if non_matching_contexts:
            raise ValueError(
                "Existing unit conversion context(s), `{}`, doesn't match input "
                "context, `{}`, drop `unit_context` metadata before doing "
                "conversion".format(unit_context, context)
            )

    def relative_to_ref_period_mean(self, append_str=None, **kwargs):
        """
        Return the timeseries relative to a given reference period mean.

        The reference period mean is subtracted from all values in the input timeseries.

        Parameters
        ----------
        append_str
            Deprecated

        **kwargs
            Arguments to pass to :func:`filter` to determine the data to be included in
            the reference time period. See the docs of :func:`filter` for valid options.

        Returns
        -------
        :obj:`ScmRun`
            New object containing the timeseries, adjusted to the reference period mean.
            The reference period year bounds are stored in the meta columns
            ``"reference_period_start_year"`` and ``"reference_period_end_year"``.

        Raises
        ------
        NotImplementedError
            ``append_str`` is not ``None``
        """
        if append_str is not None:
            raise NotImplementedError("`append_str` is deprecated")

        ts = self.timeseries()
        # mypy confused by `inplace` default
        ref_data = self.filter(**kwargs)
        ref_period_mean = ref_data.timeseries().mean(axis="columns")  # type: ignore

        res = ts.sub(ref_period_mean, axis="rows")
        res.reset_index(inplace=True)

        res["reference_period_start_year"] = ref_data["year"].min()
        res["reference_period_end_year"] = ref_data["year"].max()

        return type(self)(res)

    def append(
        self,
        other,
        inplace: bool = False,
        duplicate_msg: Union[str, bool] = True,
        metadata: Optional[MetadataType] = None,
        **kwargs: Any,
    ):
        """
        Append additional data to the current dataframe.

        For details, see :func:`run_append`.

        Parameters
        ----------
        other
            Data (in format which can be cast to :class:`ScmRun`) to append

        inplace
            If ``True``, append data in place and return ``None``. Otherwise, return a
            new :class:`ScmRun` instance with the appended data.

        duplicate_msg
            If ``True``, raise a :class:`scmdata.errors.NonUniqueMetadataError` error so the user
            can see the duplicate timeseries. If ``False``, take the average
            and do not raise a warning or error. If ``"warn"``, raise a
            warning if duplicate data is detected.

        metadata
            If not ``None``, override the metadata of the resulting :obj:`ScmRun` with ``metadata``.
            Otherwise, the metadata for the runs are merged. In the case where there are duplicate
            metadata keys, the values from the first run are used.

        **kwargs
            Keywords to pass to :func:`ScmRun.__init__` when reading
            :obj:`other`

        Returns
        -------
        :obj:`ScmRun`
            If not :obj:`inplace`, return a new :class:`ScmRun` instance
            containing the result of the append.

        Raises
        ------
        NonUniqueMetadataError
            If the appending results in timeseries with duplicate metadata and :attr:`duplicate_msg` is ``True``

        """
        if not isinstance(other, ScmRun):
            other = self.__class__(other, **kwargs)

        return run_append(
            [self, other],
            inplace=inplace,
            duplicate_msg=duplicate_msg,
            metadata=metadata,
        )

    def to_iamdataframe(self) -> LongDatetimeIamDataFrame:  # pragma: no cover
        """
        Convert to a :class:`LongDatetimeIamDataFrame` instance.

        :class:`LongDatetimeIamDataFrame` is a subclass of :class:`pyam.IamDataFrame`.
        We use :class:`LongDatetimeIamDataFrame` to ensure all times can be handled, see
        docstring of :class:`LongDatetimeIamDataFrame` for details.

        Returns
        -------
        :class:`LongDatetimeIamDataFrame`
            :class:`LongDatetimeIamDataFrame` instance containing the same data.

        Raises
        ------
        ImportError
            If `pyam <https://github.com/IAMconsortium/pyam>`_ is not installed
        """
        if LongDatetimeIamDataFrame is None:
            raise ImportError(
                "pyam is not installed. Features involving IamDataFrame are unavailable"
            )

        return LongDatetimeIamDataFrame(self.timeseries())

    def to_csv(self, fname: str, **kwargs: Any) -> None:
        """
        Write timeseries data to a csv file

        Parameters
        ----------
        fname
            Path to write the file into
        """
        self.timeseries().reset_index().to_csv(fname, **kwargs)

    def reduce(self, func, dim=None, axis=None, **kwargs):
        """
        Apply a function along a given axis

        This is to provide the GroupBy functionality in :func:`ScmRun.groupby` and is not generally called directly.

        This implementation is very bare-bones - no reduction along the time time dimension is allowed and only the `dim`
        parameter is used.

        Parameters
        ----------
        func: function
        dim : str
            Ignored
        axis : int
            The dimension along which the function is applied. The only valid value is 0 which corresponds to the along the
            time-series dimension.
        kwargs
            Other parameters passed to `func`

        Returns
        -------
        :obj:`ScmRun`

        Raises
        ------
        ValueError
            If a dimension other than None is provided

        NotImplementedError
            If `axis` is anything other than 0


        """
        if dim is not None:
            raise ValueError("ScmRun.reduce does not handle dim. Use axis instead")

        input_data = self.values

        if axis is None or axis == 1:
            raise NotImplementedError(
                "Cannot currently reduce along the time dimension"
            )

        if axis is not None:
            data = func(input_data, axis=axis, **kwargs)
        else:
            data = func(input_data, **kwargs)

        if getattr(data, "shape", ()) == self.shape:
            return type(self)(
                data, index=self.time_points, columns=self.meta.to_dict("list")
            )
        else:
            removed_axes = range(2) if axis is None else np.atleast_1d(axis) % 2
            index = self.time_points
            meta = self.meta.to_dict("list")
            if 0 in removed_axes and len(meta):
                # Reduced the timeseries
                m = self.meta
                n_unique = m.nunique(axis=0)
                m = m.drop(columns=n_unique[n_unique > 1].index).drop_duplicates()
                if len(m) != 1:  # pragma: no cover
                    raise AssertionError(m)

                meta = m.to_dict("list")

            if 1 in removed_axes:
                raise NotImplementedError  # pragma: no cover

            return type(self)(data, index=index, columns=meta)


def _merge_metadata(metadata):
    res = metadata[0].copy()

    for m in metadata[1:]:
        for k, v in m.items():
            if k not in res:
                res[k] = v
    return res


def run_append(
    runs,
    inplace: bool = False,
    duplicate_msg: Union[str, bool] = True,
    metadata: Optional[MetadataType] = None,
):
    """
    Append together many objects.

    When appending many objects, it may be more efficient to call this routine once with
    a list of :class:`ScmRun`'s, than using :func:`ScmRun.append` multiple times.

    If timeseries with duplicate metadata are found, the timeseries are appended and values
    falling on the same timestep are averaged if :obj:`duplicate_msg` is not "return". If
    :obj:`duplicate_msg` is "return", then the result will contain the duplicated timeseries
    for further inspection.

    .. code:: python

        >>> res = base.append(other, duplicate_msg="return")
        <scmdata.ScmRun (timeseries: 5, timepoints: 3)>
        Time:
            Start: 2005-01-01T00:00:00
            End: 2015-06-12T00:00:00
        Meta:
                  scenario             variable  model climate_model region   unit
            0   a_scenario       Primary Energy  a_iam       a_model  World  EJ/yr
            1   a_scenario  Primary Energy|Coal  a_iam       a_model  World  EJ/yr
            2  a_scenario2       Primary Energy  a_iam       a_model  World  EJ/yr
            3  a_scenario3       Primary Energy  a_iam       a_model  World  EJ/yr
            4   a_scenario       Primary Energy  a_iam       a_model  World  EJ/yr
        >>> ts = res.timeseries(check_duplicated=False)
        >>> ts[ts.index.duplicated(keep=False)]
        time                                                        2005-01-01  ...  2015-06-12
        scenario   variable       model climate_model region unit               ...
        a_scenario Primary Energy a_iam a_model       World  EJ/yr         1.0  ...         7.0
                                                             EJ/yr        -1.0  ...         1.0


    Parameters
    ----------
    runs: list of :class:`ScmRun`
        The runs to append. Values will be attempted to be cast to :class:`ScmRun`.

    inplace
        If ``True``, then the operation updates the first item in :obj:`runs` and returns
        ``None``.

    duplicate_msg
        If ``True``, raise a ``NonUniqueMetadataError`` error so the user can
        see the duplicate timeseries. If ``False``, take the average and do
        not raise a warning or error. If ``"warn"``, raise a warning if
        duplicate data is detected.

    metadata
        If not ``None``, override the metadata of the resulting :obj:`ScmRun` with ``metadata``.
        Otherwise, the metadata for the runs are merged. In the case where there are duplicate
        metadata keys, the values from the first run are used.

    Returns
    -------
    :obj:`ScmRun`
        If not :obj:`inplace`, the return value is the object containing the merged
        data. The resultant class will be determined by the type of the first object.

    Raises
    ------
    TypeError
        If :obj:`inplace` is ``True`` but the first element in :obj:`dfs` is not an
        instance of :class:`ScmRun`

        ``runs`` argument is not a list

    ValueError
        :obj:`duplicate_msg` option is not recognised.

        No runs are provided to be appended
    """
    if not isinstance(runs, list):
        raise TypeError("runs is not a list")

    if not len(runs):
        raise ValueError("No runs to append")

    if inplace:
        if not isinstance(runs[0], ScmRun):
            raise TypeError("Can only append inplace to an ScmRun")
        ret = runs[0]
    else:
        ret = runs[0].copy()

    to_join_dfs = []
    to_join_metas = []
    overlapping_times = False

    ind = range(ret._df.shape[1])
    ret._df.columns = ind
    ret._meta.index = ind

    min_idx = ret._df.shape[1]
    for run in runs[1:]:
        run_to_join_df = run._df

        max_idx = min_idx + run_to_join_df.shape[1]
        ind = range(min_idx, max_idx)
        min_idx = max_idx

        run_to_join_df.columns = ind
        run_to_join_meta = run._meta.to_frame()
        run_to_join_meta.index = ind

        # check everything still makes sense
        npt.assert_array_equal(run_to_join_meta.index, run_to_join_df.columns)

        # check for overlap
        idx_to_check = run_to_join_df.index
        if not overlapping_times and (
            idx_to_check.isin(ret._df.index).any()
            or any([idx_to_check.isin(df.index).any() for df in to_join_dfs])
        ):
            overlapping_times = True

        to_join_dfs.append(run_to_join_df)
        to_join_metas.append(run_to_join_meta)

    ret._df = pd.concat([ret._df] + to_join_dfs, axis="columns").sort_index()
    ret._time_points = TimePoints(ret._df.index.values)
    ret._df.index = ret._time_points.to_index()
    ret._meta = pd.MultiIndex.from_frame(
        pd.concat([ret._meta.to_frame()] + to_join_metas).astype("category")
    )

    if ret._duplicated_meta():
        if overlapping_times and duplicate_msg:
            _handle_potential_duplicates_in_append(ret, duplicate_msg)

        ts = ret.timeseries(check_duplicated=False)
        orig_ts_index = ts.index
        nan_cols = pd.isna(orig_ts_index.to_frame()).any()
        orig_dtypes = orig_ts_index.to_frame().dtypes

        # Convert index to str
        ts.index = pd.MultiIndex.from_frame(
            ts.index.to_frame().astype(str).reset_index(drop=True)
        )

        deduped_ts = ts.groupby(ts.index, as_index=True).mean()

        ret._df = deduped_ts.reset_index(drop=True).T

        new_meta = pd.DataFrame.from_records(
            deduped_ts.index.values, columns=ts.index.names
        )

        # Convert back from str
        for c in nan_cols[nan_cols].index:
            new_meta[c].replace("nan", np.nan, inplace=True)
        for c, dtype in orig_dtypes.iteritems():
            new_meta[c] = new_meta[c].astype(dtype)

        ret._meta = pd.MultiIndex.from_frame(new_meta.astype("category"))

    if metadata is not None:
        ret.metadata = metadata
    else:
        ret.metadata = _merge_metadata([r.metadata for r in runs])

    if not inplace:
        return ret


def _handle_potential_duplicates_in_append(data, duplicate_msg):
    if duplicate_msg == "warn":
        warn_msg = (
            "Duplicate time points detected, the output will be the average of "
            "the duplicates.  Set `duplicate_msg=False` to silence this message."
        )
        warnings.warn(warn_msg)
        return None

    if duplicate_msg and not isinstance(duplicate_msg, str):
        raise NonUniqueMetadataError(data.meta)

    raise ValueError("Unrecognised value for duplicate_msg")


inject_binary_ops(BaseScmRun)
inject_nc_methods(BaseScmRun)
inject_plotting_methods(BaseScmRun)
inject_ops_methods(BaseScmRun)


class ScmRun(BaseScmRun):
    """
    Data container for holding one or many time-series of SCM data.
    """

    required_cols = ("model", "scenario", "region", "variable", "unit")
    """
    Minimum metadata columns required by an ScmRun.

    If an application requires a different set of required metadata, this
    can be specified by overriding :attr:`required_cols` on a custom class
    inheriting :class:`scmdata.run.BaseScmRun`. Note that at a minimum,
    ("variable", "unit") columns are required.
    """



"""
Imports and classes required to ensure compatibility with Pyam is intelligently
handled.
"""
import datetime

from dateutil import parser

try:
    from pyam import IamDataFrame

    # mypy can't work out try-except block forces IamDataFrame to be here
    class LongDatetimeIamDataFrame(IamDataFrame):  # type: ignore
        """
        Custom implementation of :class:`pyam.IamDataFrame` which handles long datetime
        data.

        This custom implementation allows the data frame to handle times outside panda's
        limited time range of 1677-09-22 00:12:43.145225 to 2262-04-11
        23:47:16.854775807, see `this discussion
        <https://stackoverflow.com/a/37226672>`_.

        TODO: use np.datetime64 instead
        """

        def _format_datetime_col(self):
            if self.data["time"].apply(lambda x: isinstance(x, str)).any():

                def convert_str_to_datetime(inp):
                    try:
                        return parser.parse(inp)
                    except ValueError:
                        return inp

                self.data["time"] = self.data["time"].apply(convert_str_to_datetime)

            not_datetime = [
                not isinstance(x, (datetime.datetime, datetime.date))
                for x in self.data["time"]
            ]
            if any(not_datetime):
                bad_values = self.data[not_datetime]["time"]
                error_msg = "All time values must be convertible to datetime. The following values are not:\n{}".format(
                    bad_values
                )
                raise ValueError(error_msg)


except ImportError:  # pragma: no cover
    # mypy can't work out try-except block sets typing
    IamDataFrame = None
    LongDatetimeIamDataFrame = None  # type: ignore



"""
NetCDF4 file operations

Reading and writing :obj:`ScmRun` to disk as binary
"""
try:
    import netCDF4 as nc

    has_netcdf = True
except ImportError:  # pragma: no cover
    nc = None
    has_netcdf = False

from collections import defaultdict
from datetime import datetime
from logging import getLogger

import numpy as np

from . import __version__

logger = getLogger(__name__)

_TO_NC_DOCSTRING = """\
Write data to disk as a netCDF4 file

Parameters
----------
path: str
    Path to write the file into

dimensions: iterable of str
    Dimensions to include in the netCDF file. The order of the dimensions in the netCDF file will be the same
    as the order provided.
    The time dimension is always included as the last dimension, even if not provided.
"""


_FROM_NC_DOCSTRING = """\
Read netCDF4 file from disk

Parameters
----------
path: str
    Path to write the file into

"""

"""
Default to writing float data as 8 byte floats
"""
DEFAULT_FLOAT = "f8"


def _var_to_nc(var):
    return var.replace("|", "__").replace(" ", "_")


def _nc_to_var(var):
    return var.replace("__", "|").replace("_", " ")


def _get_idx(vals, v):
    if v not in vals:
        raise AssertionError("{} is not in {}".format(v, vals))

    return np.where(vals == v)[0][0]


def _get_nc_type(np_type):
    if np_type == int:
        return {
            "datatype": "i8",
        }
    elif np_type == float:
        return {"datatype": DEFAULT_FLOAT, "fill_value": np.nan}

    return {"datatype": str, "fill_value": None}


def _write_nc(ds, df, dimensions, extras):
    """
    Low level function to write the dimensions, variables and metadata to disk
    """
    all_dims = list(dimensions) + ["time"]

    # Create the dimensions
    ds.createDimension("time", len(df.time_points))
    ds.createVariable(
        "time", "f8", "time",
    )
    ds.variables["time"][:] = df.time_points.values

    dims = {}
    for d in dimensions:
        vals = sorted(df.meta[d].unique())
        if not all([isinstance(v, str) for v in vals]) and np.isnan(vals).any():
            raise AssertionError("nan in dimension: `{}`".format(d))

        ds.createDimension(d, len(vals))
        dtype = type(vals[0])
        ds.createVariable(d, dtype, d)
        for i, v in enumerate(vals):  # Iteration needed for str types
            ds.variables[d][i] = v
        dims[d] = np.asarray(vals)

    var_shape = [len(dims[d]) for d in dimensions] + [len(df.time_points)]

    # Write any extra variables
    for e in extras:
        metadata = df.meta[[e, *dimensions]].drop_duplicates()

        if metadata[dimensions].duplicated().any():
            raise ValueError(
                "metadata for {} is not unique for requested dimensions".format(e)
            )

        type_info = _get_nc_type(metadata[e].dtype)
        ds.createVariable(e, dimensions=dimensions, zlib=True, **type_info)
        ds.variables[e]._is_metadata = 1

        data_to_write = np.zeros(
            [len(dims[d]) for d in dimensions], dtype=metadata[e].dtype
        )
        if "fill_value" in type_info:
            data_to_write.fill(type_info["fill_value"])

        df_values = metadata[e].values
        for i, (_, m) in enumerate(metadata.iterrows()):
            idx = [_get_idx(dims[d], m[d]) for d in dimensions]
            data_to_write[tuple(idx)] = df_values[i]

        ds.variables[e][:] = data_to_write

    for var_df in df.groupby("variable"):
        v = var_df.get_unique_meta("variable", True)
        meta = var_df.meta.copy().drop("variable", axis=1)

        # Check that the varying dimensions are all unique
        if meta[dimensions].duplicated().any():
            raise ValueError(
                "{} dimensions are not unique for variable {}".format(dimensions, v)
            )

        # Check that the other meta are consistent
        var_attrs = {"_is_metadata": 0}
        for d in set(meta.columns) - set(dimensions) - set(extras):
            if len(meta[d].unique()) != 1:
                raise ValueError(
                    "metadata for {} is not unique for variable {}".format(d, v)
                )
            var_attrs[d] = meta[d].unique()[0]

        var_name = _var_to_nc(v)
        ds.createVariable(
            var_name, DEFAULT_FLOAT, all_dims, zlib=True, fill_value=np.nan
        )

        # We need to write in one dimension at a time
        data_to_write = np.zeros(var_shape)
        data_to_write.fill(np.nan)
        df_values = var_df.values
        for i, (_, m) in enumerate(meta.iterrows()):
            idx = [_get_idx(dims[d], m[d]) for d in dimensions]
            idx.append(slice(None))  # time dim
            data_to_write[tuple(idx)] = df_values[i]
        # Write in one call to the nc library
        ds.variables[var_name][:] = data_to_write

        # Set variable metadata
        ds.variables[var_name].setncatts(var_attrs)


def _read_nc(cls, ds):
    dims = {d: ds.variables[d][:] for d in ds.dimensions}
    dims["time"] = dims["time"].astype("datetime64[s]")

    data = []
    columns = defaultdict(list)

    def _read_var(name, var):
        var_data = var[:]
        valid_mask = ~np.isnan(var_data).all(axis=-1)

        var_meta = {"variable": name}
        for v in var.ncattrs():
            if not v.startswith("_"):
                var_meta[v] = var.getncattr(v)

        # Iterate over all combinations of dimensions
        if len(var.dimensions) > 1:
            meta_at_coord = np.asarray(
                np.meshgrid(*[dims[d] for d in var.dimensions[:-1]], indexing="ij")
            )
            meta_at_coord = meta_at_coord[0]

            with np.nditer(meta_at_coord, ["refs_ok", "multi_index"], order="F") as it:
                for _ in it:
                    if not valid_mask[it.multi_index]:
                        continue
                    data.append(var_data[it.multi_index])
                    for i, v in enumerate(it.multi_index):
                        dim_name = var.dimensions[i]
                        columns[dim_name].append(dims[dim_name][v])
                    for v in var_meta:
                        columns[v].append(var_meta[v])
        else:
            data.append(var_data[:])
            for v in var_meta:
                columns[v].append(var_meta[v])

    extra_cols = []
    for var_name in ds.variables:
        var = ds.variables[var_name]
        if var_name in dims:
            continue

        # Check if metadata column
        if var.getncattr("_is_metadata"):
            extra_cols.append(var_name)

            # Intialise the required columns and load later
            if var_name in cls.required_cols:
                columns[var_name] = ""
            continue

        name = _nc_to_var(var_name)
        _read_var(name, var)

    df = cls(np.asarray(data).T, columns=columns, index=dims["time"])

    # Parse any extra metadata columns
    # Requires 1 filter per item

    def _merge_meta(meta, col, value, **filters):
        result = []
        for i, r in meta.iterrows():
            matches = True
            val = r[col]
            for k, v in filters.items():
                if r[k] != v:
                    matches = False
                    break
            if matches:
                val = value
            result.append(val)
        return result

    for col in extra_cols:
        var = ds.variables[col]
        df[col] = None

        values = var[:]
        if len(var.dimensions):
            meta_at_coord = np.asarray(
                np.meshgrid(*[dims[d] for d in var.dimensions], indexing="ij")
            )
            with np.nditer(
                meta_at_coord[0], ["refs_ok", "multi_index"], order="F"
            ) as it:
                for _ in it:
                    meta_vals = {
                        k: v
                        for k, v in zip(
                            var.dimensions,
                            meta_at_coord[(slice(None),) + it.multi_index],
                        )
                    }
                    df[col] = _merge_meta(
                        df.meta, col, values[it.multi_index], **meta_vals
                    )
        else:
            df[col] = var[:]
        try:
            df[col] = df[col].astype(var.dtype)
        except ValueError:
            logger.exception("could not convert meta dtype")

    df.metadata.update({k: ds.getncattr(k) for k in ds.ncattrs()})

    return df


def run_to_nc(run, fname, dimensions=("region",), extras=()):
    """
    Write timeseries to disk as a netCDF4 file

    Each unique variable will be written as a netCDF file.

    Parameters
    ----------
    fname: str
        Path to write the file into

    dimensions: iterable of str
        Dimensions to include in the netCDF file. The order of the dimensions
        in the netCDF file will be the same as the order provided. The time
        dimension is always included as the last dimension, even if not provided.

    extras : iterable of tuples or str
        Metadata attributes to write as variables in the netCDF file.

    See Also
    --------
    :meth:`scmdata.run.ScmRun.to_nc`
    """
    if not has_netcdf:
        raise ImportError("netcdf4 is not installed. Run 'pip install netcdf4'")

    dimensions = list(dimensions)
    if "time" in dimensions:
        dimensions.remove("time")
    if "variable" in dimensions:
        dimensions.remove("variable")

    with nc.Dataset(fname, "w", diskless=True, persist=True) as ds:
        ds.created_at = datetime.utcnow().isoformat()
        ds._scmdata_version = __version__

        if hasattr(run, "metadata"):
            ds.setncatts(run.metadata)

        _write_nc(ds, run, dimensions, extras)


def nc_to_run(cls, fname):
    """
    Read a netCDF4 file from disk

    Parameters
    ----------
    fname: str
        Filename to read

    See Also
    --------
    :meth:`scmdata.run.ScmRun.from_nc`
    """
    if not has_netcdf:
        raise ImportError("netcdf4 is not installed. Run 'pip install netcdf4'")

    with nc.Dataset(fname) as ds:
        try:
            return _read_nc(cls, ds)
        except Exception:
            logger.exception("Failed reading netcdf file: {}".format(fname))
            raise


def inject_nc_methods(cls):
    """
    Add the to/from nc methods to a class

    Parameters
    ----------
    cls
        Class to add methods to
    """
    name = "to_nc"
    func = run_to_nc
    func.__name__ = name
    func.__doc__ = func.__doc__
    setattr(cls, name, func)

    name = "from_nc"
    func = classmethod(nc_to_run)
    func.__name__ = name
    func.__doc__ = func.__doc__
    setattr(cls, name, func)



"""
Operations for :obj:`ScmRun`

These largely rely on
`Pint's Pandas interface <https://pint.readthedocs.io/en/0.13/pint-pandas.html>`_
to handle unit conversions automatically
"""
import warnings

import numpy as np
import pandas as pd
import pint_pandas
from openscm_units import unit_registry

from .time import TimePoints

try:
    import scipy.integrate

    has_scipy = True
except ImportError:  # pragma: no cover
    scipy = None
    has_scipy = False


def prep_for_op(inp, op_cols, meta, ur=unit_registry):
    """
    Prepare dataframe for operation

    Parameters
    ----------
    inp : :obj:`ScmRun`
        :obj:`ScmRun` containing data to prepare

    op_cols : dict of str: str
        Dictionary containing the columns to drop in order to prepare
        for the operation as the keys (the values are not used). For
        example, if we have
        ``op_cols={"variable": "Emissions|CO2 - Emissions|CO2|Fossil"}``
        then we will drop the "variable" column from the index.

    ur : :obj:`pint.UnitRegistry`
        Pint unit registry to use for the operation

    Returns
    -------
    :obj:`pd.DataFrame`
        Timeseries to use for the operation. They are the transpose of
        the normal :meth:`ScmRun.timeseries` output with the columns
        being Pint arrays (unless "unit" is in op_cols in which case no
        units are available to be used so the columns are standard numpy
        arrays). We do this so that we can use
        `Pint's Pandas interface <https://pint.readthedocs.io/en/0.13/pint-pandas.html>`_
        to handle unit conversions automatically.
    """
    pint_pandas.PintType.ureg = ur

    key_cols = list(op_cols.keys())

    out = inp.timeseries(meta=meta).reset_index(key_cols, drop=True)

    out = out.T

    if "unit" not in op_cols:
        unit_level = out.columns.names.index("unit")
        out = out.pint.quantify(level=unit_level)

    return out


def set_op_values(output, op_cols):
    """
    Set operation values in output

    Parameters
    ----------
    output : :obj:`pd.Dataframe`
        Dataframe of which to update the values

    op_cols : dict of str: str
        Dictionary containing the columns to update as the keys and the value
        those columns should hold in the output as the values. For example,
        if we have
        ``op_cols={"variable": "Emissions|CO2 - Emissions|CO2|Fossil"}`` then
        the output will have a "variable" column with the value
        "Emissions|CO2 - Emissions|CO2|Fossil".

    Returns
    -------
    :obj:`pd.Dataframe`
        ``output`` with the relevant columns being set according to ``op_cols``.
    """
    for k, v in op_cols.items():
        output[k] = v

    return output


def _perform_op(base, other, op, use_pint_units=True):
    # pint handling means we have to do series by series
    out = []
    col_names = base.columns.names
    for col in base:
        if op in ["add", "subtract"]:
            try:
                if op == "add":
                    out.append(base[col] + other[col])

                elif op == "subtract":
                    out.append(base[col] - other[col])

            except KeyError:
                raise KeyError(
                    "No equivalent in `other` for {}".format(list(zip(col_names, col)))
                )

        elif op == "multiply":
            out.append(base[col] * other[col])

        elif op == "divide":
            out.append(base[col] / other[col])

        else:  # pragma: no cover
            raise NotImplementedError(op)

    if len(out) == 1:
        out = out[0].to_frame()
    else:
        out = pd.concat(out, axis="columns")

    out.columns.names = base.columns.names

    if use_pint_units:
        out = out.pint.dequantify()

    out = out.T

    return out


def subtract(self, other, op_cols, **kwargs):
    """
    Subtract values

    Parameters
    ----------
    other : :obj:`ScmRun`
        :obj:`ScmRun` containing data to subtract

    op_cols : dict of str: str
        Dictionary containing the columns to drop before subtracting as
        the keys and the value those columns should hold in the output
        as the values. For example, if we have
        ``op_cols={"variable": "Emissions|CO2 - Emissions|CO2|Fossil"}``
        then the subtraction will be performed with an index that uses
        all columns except the "variable" column and the output will
        have a "variable" column with the value
        "Emissions|CO2 - Emissions|CO2|Fossil".

    **kwargs : any
        Passed to :func:`prep_for_op`

    Returns
    -------
    :obj:`ScmRun`
        Difference between ``self`` and ``other``, using ``op_cols`` to
        define the columns which should be dropped before the data is
        aligned and to define the value of these columns in the output.

    Examples
    --------
        >>> import numpy as np
    >>> from scmdata import ScmRun
    >>>
    >>> IDX = [2010, 2020, 2030]
    >>>
    >>>
    >>> start = ScmRun(
    ...     data=np.arange(18).reshape(3, 6),
    ...     index=IDX,
    ...     columns={
    ...         "variable": [
    ...             "Emissions|CO2|Fossil",
    ...             "Emissions|CO2|AFOLU",
    ...             "Emissions|CO2|Fossil",
    ...             "Emissions|CO2|AFOLU",
    ...             "Cumulative Emissions|CO2",
    ...             "Surface Air Temperature Change",
    ...         ],
    ...         "unit": ["GtC / yr", "GtC / yr", "GtC / yr", "GtC / yr", "GtC", "K"],
    ...         "region": ["World|NH", "World|NH", "World|SH", "World|SH", "World", "World"],
    ...         "model": "idealised",
    ...         "scenario": "idealised",
    ...     },
    ... )
    >>>
    >>> start.head()
    time                                                            2010-01-01 00:00:00  2020-01-01 00:00:00  2030-01-01 00:00:00
    variable                 unit     region   model     scenario
    Emissions|CO2|Fossil     GtC / yr World|NH idealised idealised                  0.0                  6.0                 12.0
    Emissions|CO2|AFOLU      GtC / yr World|NH idealised idealised                  1.0                  7.0                 13.0
    Emissions|CO2|Fossil     GtC / yr World|SH idealised idealised                  2.0                  8.0                 14.0
    Emissions|CO2|AFOLU      GtC / yr World|SH idealised idealised                  3.0                  9.0                 15.0
    Cumulative Emissions|CO2 GtC      World    idealised idealised                  4.0                 10.0                 16.0
    >>> fos = start.filter(variable="*Fossil")
    >>> fos.head()
    time                                                        2010-01-01 00:00:00  2020-01-01 00:00:00  2030-01-01 00:00:00
    variable             unit     region   model     scenario
    Emissions|CO2|Fossil GtC / yr World|NH idealised idealised                  0.0                  6.0                 12.0
                                  World|SH idealised idealised                  2.0                  8.0                 14.0
    >>>
    >>> afolu = start.filter(variable="*AFOLU")
    >>> afolu.head()
    time                                                       2010-01-01 00:00:00  2020-01-01 00:00:00  2030-01-01 00:00:00
    variable            unit     region   model     scenario
    Emissions|CO2|AFOLU GtC / yr World|NH idealised idealised                  1.0                  7.0                 13.0
                                 World|SH idealised idealised                  3.0                  9.0                 15.0
    >>>
    >>> fos_minus_afolu = fos.subtract(
    ...     afolu, op_cols={"variable": "Emissions|CO2|Fossil - AFOLU"}
    ... )
    >>> fos_minus_afolu.head()
    time                                                                  2010-01-01 00:00:00  2020-01-01 00:00:00  2030-01-01 00:00:00
    model     scenario  region   variable                     unit
    idealised idealised World|NH Emissions|CO2|Fossil - AFOLU gigatC / a                 -1.0                 -1.0                 -1.0
                        World|SH Emissions|CO2|Fossil - AFOLU gigatC / a                 -1.0                 -1.0                 -1.0
    >>>
    >>> nh_minus_sh = nh.subtract(sh, op_cols={"region": "World|NH - SH"})
    >>> nh_minus_sh.head()
    time                                                               2010-01-01 00:00:00  2020-01-01 00:00:00  2030-01-01 00:00:00
    model     scenario  region        variable             unit
    idealised idealised World|NH - SH Emissions|CO2|Fossil gigatC / a                 -2.0                 -2.0                 -2.0
                                      Emissions|CO2|AFOLU  gigatC / a                 -2.0                 -2.0                 -2.0
    """
    out = _perform_op(
        prep_for_op(self, op_cols, self.meta.columns, **kwargs),
        prep_for_op(other, op_cols, self.meta.columns, **kwargs),
        "subtract",
        use_pint_units="unit" not in op_cols,
    )

    out = set_op_values(out, op_cols)

    return type(self)(out)


def add(self, other, op_cols, **kwargs):
    """
    Add values

    Parameters
    ----------
    other : :obj:`ScmRun`
        :obj:`ScmRun` containing data to add

    op_cols : dict of str: str
        Dictionary containing the columns to drop before adding as the
        keys and the value those columns should hold in the output as
        the values. For example, if we have
        ``op_cols={"variable": "Emissions|CO2 - Emissions|CO2|Fossil"}``
        then the addition will be performed with an index that uses
        all columns except the "variable" column and the output will
        have a "variable" column with the value
        "Emissions|CO2 - Emissions|CO2|Fossil".

    **kwargs : any
        Passed to :func:`prep_for_op`

    Returns
    -------
    :obj:`ScmRun`
        Sum of ``self`` and ``other``, using ``op_cols`` to define the
        columns which should be dropped before the data is aligned and
        to define the value of these columns in the output.

    Examples
    --------
    >>> import numpy as np
    >>> from scmdata import ScmRun
    >>>
    >>> IDX = [2010, 2020, 2030]
    >>>
    >>>
    >>> start = ScmRun(
    ...     data=np.arange(18).reshape(3, 6),
    ...     index=IDX,
    ...     columns={
    ...         "variable": [
    ...             "Emissions|CO2|Fossil",
    ...             "Emissions|CO2|AFOLU",
    ...             "Emissions|CO2|Fossil",
    ...             "Emissions|CO2|AFOLU",
    ...             "Cumulative Emissions|CO2",
    ...             "Surface Air Temperature Change",
    ...         ],
    ...         "unit": ["GtC / yr", "GtC / yr", "GtC / yr", "GtC / yr", "GtC", "K"],
    ...         "region": ["World|NH", "World|NH", "World|SH", "World|SH", "World", "World"],
    ...         "model": "idealised",
    ...         "scenario": "idealised",
    ...     },
    ... )
    >>>
    >>> start.head()
    time                                                            2010-01-01 00:00:00  2020-01-01 00:00:00  2030-01-01 00:00:00
    variable                 unit     region   model     scenario
    Emissions|CO2|Fossil     GtC / yr World|NH idealised idealised                  0.0                  6.0                 12.0
    Emissions|CO2|AFOLU      GtC / yr World|NH idealised idealised                  1.0                  7.0                 13.0
    Emissions|CO2|Fossil     GtC / yr World|SH idealised idealised                  2.0                  8.0                 14.0
    Emissions|CO2|AFOLU      GtC / yr World|SH idealised idealised                  3.0                  9.0                 15.0
    Cumulative Emissions|CO2 GtC      World    idealised idealised                  4.0                 10.0                 16.0
    >>> fos = start.filter(variable="*Fossil")
    >>> fos.head()
    time                                                        2010-01-01 00:00:00  2020-01-01 00:00:00  2030-01-01 00:00:00
    variable             unit     region   model     scenario
    Emissions|CO2|Fossil GtC / yr World|NH idealised idealised                  0.0                  6.0                 12.0
                                  World|SH idealised idealised                  2.0                  8.0                 14.0
    >>>
    >>> afolu = start.filter(variable="*AFOLU")
    >>> afolu.head()
    time                                                       2010-01-01 00:00:00  2020-01-01 00:00:00  2030-01-01 00:00:00
    variable            unit     region   model     scenario
    Emissions|CO2|AFOLU GtC / yr World|NH idealised idealised                  1.0                  7.0                 13.0
                                 World|SH idealised idealised                  3.0                  9.0                 15.0
    >>>
    >>> total = fos.add(afolu, op_cols={"variable": "Emissions|CO2"})
    >>> total.head()
    time                                                   2010-01-01 00:00:00  2020-01-01 00:00:00  2030-01-01 00:00:00
    model     scenario  region   variable      unit
    idealised idealised World|NH Emissions|CO2 gigatC / a                  1.0                 13.0                 25.0
                        World|SH Emissions|CO2 gigatC / a                  5.0                 17.0                 29.0
    >>>
    >>> nh = start.filter(region="*NH")
    >>> nh.head()
    time                                                        2010-01-01 00:00:00  2020-01-01 00:00:00  2030-01-01 00:00:00
    variable             unit     region   model     scenario
    Emissions|CO2|Fossil GtC / yr World|NH idealised idealised                  0.0                  6.0                 12.0
    Emissions|CO2|AFOLU  GtC / yr World|NH idealised idealised                  1.0                  7.0                 13.0
    >>>
    >>> sh = start.filter(region="*SH")
    >>> sh.head()
    time                                                        2010-01-01 00:00:00  2020-01-01 00:00:00  2030-01-01 00:00:00
    variable             unit     region   model     scenario
    Emissions|CO2|Fossil GtC / yr World|SH idealised idealised                  2.0                  8.0                 14.0
    Emissions|CO2|AFOLU  GtC / yr World|SH idealised idealised                  3.0                  9.0                 15.0
    >>>
    >>> world = nh.add(sh, op_cols={"region": "World"})
    >>> world.head()
    time                                                        2010-01-01 00:00:00  2020-01-01 00:00:00  2030-01-01 00:00:00
    model     scenario  region variable             unit
    idealised idealised World  Emissions|CO2|Fossil gigatC / a                  2.0                 14.0                 26.0
                               Emissions|CO2|AFOLU  gigatC / a                  4.0                 16.0                 28.0
    """
    out = _perform_op(
        prep_for_op(self, op_cols, self.meta.columns, **kwargs),
        prep_for_op(other, op_cols, self.meta.columns, **kwargs),
        "add",
        use_pint_units="unit" not in op_cols,
    )

    out = set_op_values(out, op_cols)

    return type(self)(out)


def multiply(self, other, op_cols, **kwargs):
    """
    Multiply values

    Parameters
    ----------
    other : :obj:`ScmRun`
        :obj:`ScmRun` containing data to multiply

    op_cols : dict of str: str
        Dictionary containing the columns to drop before multiplying as the keys and the value those columns should hold in the output as the values. For example, if we have ``op_cols={"variable": "Emissions|CO2 - Emissions|CO2|Fossil"}`` then the multiplication will be performed with an index that uses all columns except the "variable" column and the output will have a "variable" column with the value "Emissions|CO2 - Emissions|CO2|Fossil".

    **kwargs : any
        Passed to :func:`prep_for_op`

    Returns
    -------
    :obj:`ScmRun`
        Product of ``self`` and ``other``, using ``op_cols`` to define the columns which should be dropped before the data is aligned and to define the value of these columns in the output.

    Examples
    --------
    >>> import numpy as np
    >>> from scmdata import ScmRun
    >>>
    >>> IDX = [2010, 2020, 2030]
    >>>
    >>>
    >>> start = ScmRun(
    ...     data=np.arange(18).reshape(3, 6),
    ...     index=IDX,
    ...     columns={
    ...         "variable": [
    ...             "Emissions|CO2|Fossil",
    ...             "Emissions|CO2|AFOLU",
    ...             "Emissions|CO2|Fossil",
    ...             "Emissions|CO2|AFOLU",
    ...             "Cumulative Emissions|CO2",
    ...             "Surface Air Temperature Change",
    ...         ],
    ...         "unit": ["GtC / yr", "GtC / yr", "GtC / yr", "GtC / yr", "GtC", "K"],
    ...         "region": ["World|NH", "World|NH", "World|SH", "World|SH", "World", "World"],
    ...         "model": "idealised",
    ...         "scenario": "idealised",
    ...     },
    ... )
    >>>
    >>> start.head()
    time                                                            2010-01-01 00:00:00  2020-01-01 00:00:00  2030-01-01 00:00:00
    variable                 unit     region   model     scenario
    Emissions|CO2|Fossil     GtC / yr World|NH idealised idealised                  0.0                  6.0                 12.0
    Emissions|CO2|AFOLU      GtC / yr World|NH idealised idealised                  1.0                  7.0                 13.0
    Emissions|CO2|Fossil     GtC / yr World|SH idealised idealised                  2.0                  8.0                 14.0
    Emissions|CO2|AFOLU      GtC / yr World|SH idealised idealised                  3.0                  9.0                 15.0
    Cumulative Emissions|CO2 GtC      World    idealised idealised                  4.0                 10.0                 16.0
    >>> fos = start.filter(variable="*Fossil")
    >>> fos.head()
    time                                                        2010-01-01 00:00:00  2020-01-01 00:00:00  2030-01-01 00:00:00
    variable             unit     region   model     scenario
    Emissions|CO2|Fossil GtC / yr World|NH idealised idealised                  0.0                  6.0                 12.0
                                  World|SH idealised idealised                  2.0                  8.0                 14.0
    >>>
    >>> afolu = start.filter(variable="*AFOLU")
    >>> afolu.head()
    time                                                       2010-01-01 00:00:00  2020-01-01 00:00:00  2030-01-01 00:00:00
    variable            unit     region   model     scenario
    Emissions|CO2|AFOLU GtC / yr World|NH idealised idealised                  1.0                  7.0                 13.0
                                 World|SH idealised idealised                  3.0                  9.0                 15.0
    >>>
    >>> fos_times_afolu = fos.multiply(
    ...     afolu, op_cols={"variable": "Emissions|CO2|Fossil : AFOLU"}
    ... )
    >>> fos_times_afolu.head()
    time                                                                            2010-01-01 00:00:00  2020-01-01 00:00:00  2030-01-01 00:00:00
    model     scenario  region   variable                     unit
    idealised idealised World|NH Emissions|CO2|Fossil : AFOLU gigatC ** 2 / a ** 2                  0.0                 42.0                156.0
                        World|SH Emissions|CO2|Fossil : AFOLU gigatC ** 2 / a ** 2                  6.0                 72.0                210.0
    """
    out = _perform_op(
        prep_for_op(self, op_cols, self.meta.columns, **kwargs),
        prep_for_op(other, op_cols, self.meta.columns, **kwargs),
        "multiply",
        use_pint_units="unit" not in op_cols,
    )

    out = set_op_values(out, op_cols)

    return type(self)(out)


def divide(self, other, op_cols, **kwargs):
    """
    Divide values (self / other)

    Parameters
    ----------
    other : :obj:`ScmRun`
        :obj:`ScmRun` containing data to divide

    op_cols : dict of str: str
        Dictionary containing the columns to drop before dividing as the keys and the value those columns should hold in the output as the values. For example, if we have ``op_cols={"variable": "Emissions|CO2 - Emissions|CO2|Fossil"}`` then the division will be performed with an index that uses all columns except the "variable" column and the output will have a "variable" column with the value "Emissions|CO2 - Emissions|CO2|Fossil".

    **kwargs : any
        Passed to :func:`prep_for_op`

    Returns
    -------
    :obj:`ScmRun`
        Quotient of ``self`` and ``other``, using ``op_cols`` to define the columns which should be dropped before the data is aligned and to define the value of these columns in the output.

    Examples
    --------
    >>> import numpy as np
    >>> from scmdata import ScmRun
    >>>
    >>> IDX = [2010, 2020, 2030]
    >>>
    >>>
    >>> start = ScmRun(
    ...     data=np.arange(18).reshape(3, 6),
    ...     index=IDX,
    ...     columns={
    ...         "variable": [
    ...             "Emissions|CO2|Fossil",
    ...             "Emissions|CO2|AFOLU",
    ...             "Emissions|CO2|Fossil",
    ...             "Emissions|CO2|AFOLU",
    ...             "Cumulative Emissions|CO2",
    ...             "Surface Air Temperature Change",
    ...         ],
    ...         "unit": ["GtC / yr", "GtC / yr", "GtC / yr", "GtC / yr", "GtC", "K"],
    ...         "region": ["World|NH", "World|NH", "World|SH", "World|SH", "World", "World"],
    ...         "model": "idealised",
    ...         "scenario": "idealised",
    ...     },
    ... )
    >>>
    >>> start.head()
    time                                                            2010-01-01 00:00:00  2020-01-01 00:00:00  2030-01-01 00:00:00
    variable                 unit     region   model     scenario
    Emissions|CO2|Fossil     GtC / yr World|NH idealised idealised                  0.0                  6.0                 12.0
    Emissions|CO2|AFOLU      GtC / yr World|NH idealised idealised                  1.0                  7.0                 13.0
    Emissions|CO2|Fossil     GtC / yr World|SH idealised idealised                  2.0                  8.0                 14.0
    Emissions|CO2|AFOLU      GtC / yr World|SH idealised idealised                  3.0                  9.0                 15.0
    Cumulative Emissions|CO2 GtC      World    idealised idealised                  4.0                 10.0                 16.0
    >>> fos = start.filter(variable="*Fossil")
    >>> fos.head()
    time                                                        2010-01-01 00:00:00  2020-01-01 00:00:00  2030-01-01 00:00:00
    variable             unit     region   model     scenario
    Emissions|CO2|Fossil GtC / yr World|NH idealised idealised                  0.0                  6.0                 12.0
                                  World|SH idealised idealised                  2.0                  8.0                 14.0
    >>>
    >>> afolu = start.filter(variable="*AFOLU")
    >>> afolu.head()
    time                                                       2010-01-01 00:00:00  2020-01-01 00:00:00  2030-01-01 00:00:00
    variable            unit     region   model     scenario
    Emissions|CO2|AFOLU GtC / yr World|NH idealised idealised                  1.0                  7.0                 13.0
                                 World|SH idealised idealised                  3.0                  9.0                 15.0
    >>>
    >>> fos_afolu_ratio = fos.divide(
    ...     afolu, op_cols={"variable": "Emissions|CO2|Fossil : AFOLU"}
    ... )
    >>> fos_afolu_ratio.head()
    time                                                                     2010-01-01 00:00:00  2020-01-01 00:00:00  2030-01-01 00:00:00
    model     scenario  region   variable                     unit
    idealised idealised World|NH Emissions|CO2|Fossil : AFOLU dimensionless             0.000000             0.857143             0.923077
                        World|SH Emissions|CO2|Fossil : AFOLU dimensionless             0.666667             0.888889             0.933333
    """
    out = _perform_op(
        prep_for_op(self, op_cols, self.meta.columns, **kwargs),
        prep_for_op(other, op_cols, self.meta.columns, **kwargs),
        "divide",
        use_pint_units="unit" not in op_cols,
    )

    out = set_op_values(out, op_cols)

    return type(self)(out)


def integrate(self, out_var=None):
    """
    Integrate with respect to time

    Parameters
    ----------
    out_var : str
        If provided, the variable column of the output is set equal to
        ``out_var``. Otherwise, the output variables are equal to the input
        variables, prefixed with "Cumulative " .

    Returns
    -------
    :obj:`scmdata.ScmRun`
        :obj:`scmdata.ScmRun` containing the integral of ``self`` with respect
        to time

    Warns
    -----
    UserWarning
        The data being integrated contains nans. If this happens, the output
        data will also contain nans.
    """
    if not has_scipy:
        raise ImportError("scipy is not installed. Run 'pip install scipy'")

    time_unit = "s"
    times_in_s = self.time_points.values.astype(
        "datetime64[{}]".format(time_unit)
    ).astype("int")

    ts = self.timeseries()
    if ts.isnull().sum().sum() > 0:
        warnings.warn(
            "You are integrating data which contains nans so your result will "
            "also contain nans. Perhaps you want to remove the nans before "
            "performing the integration using a combination of :meth:`filter` "
            "and :meth:`interpolate`?"
        )
    # If required, we can remove the hard-coding of initial, it just requires
    # some thinking about unit handling
    _initial = 0.0
    out = pd.DataFrame(
        scipy.integrate.cumtrapz(y=ts, x=times_in_s, axis=1, initial=_initial)
    )
    out.index = ts.index
    out.columns = ts.columns

    out = type(self)(out)
    out *= unit_registry(time_unit)

    try:
        out_unit = out.get_unique_meta("unit", no_duplicates=True).replace(" ", "")
        out_unit = str(unit_registry(out_unit).to_reduced_units().units)
        out = out.convert_unit(out_unit)

    except ValueError:
        # more than one unit, don't try to clean up
        pass

    if out_var is None:
        out["variable"] = "Cumulative " + out["variable"]
    else:
        out["variable"] = out_var

    return out


def delta_per_delta_time(self, out_var=None):
    """
    Calculate change in timeseries values for each timestep, divided by the size of the timestep

    The output is placed on the middle of each timestep and is one timestep
    shorter than the input.

    Parameters
    ----------
    out_var : str
        If provided, the variable column of the output is set equal to
        ``out_var``. Otherwise, the output variables are equal to the input
        variables, prefixed with "Delta " .

    Returns
    -------
    :obj:`scmdata.ScmRun`
        :obj:`scmdata.ScmRun` containing the changes in values of ``self``,
        normalised by the change in time

    Warns
    -----
    UserWarning
        The data contains nans. If this happens, the output data will also
        contain nans.
    """
    time_unit = "s"
    times_numpy = self.time_points.values.astype("datetime64[{}]".format(time_unit))
    times_deltas_numpy = times_numpy[1:] - times_numpy[:-1]
    times_in_s = times_numpy.astype("int")
    time_deltas_in_s = times_in_s[1:] - times_in_s[:-1]

    ts = self.timeseries()
    if ts.isnull().sum().sum() > 0:
        warnings.warn(
            "You are calculating deltas of data which contains nans so your "
            "result will also contain nans. Perhaps you want to remove the "
            "nans before calculating the deltas using a combination of "
            ":meth:`filter` and :meth:`interpolate`?"
        )

    out = ts.diff(periods=1, axis="columns")
    if not out.iloc[:, 0].isnull().all():  # pragma: no cover
        raise AssertionError(
            "Did pandas change their API? The first timestep is not all nan."
        )

    out = out.iloc[:, 1:] / time_deltas_in_s

    new_times = times_numpy[:-1] + times_deltas_numpy / 2
    out.columns = TimePoints(new_times).to_index()

    out = type(self)(out)
    out /= unit_registry(time_unit)

    try:
        out_unit = out.get_unique_meta("unit", no_duplicates=True).replace(" ", "")
        out_unit = str(unit_registry(out_unit).to_reduced_units().units)
        out = out.convert_unit(out_unit)

    except ValueError:
        # more than one unit, don't try to clean up
        pass

    if out_var is None:
        out["variable"] = "Delta " + out["variable"]
    else:
        out["variable"] = out_var

    return out


def linear_regression(self):
    """
    Calculate linear regression of each timeseries

    Note
    ----
    Times in seconds since 1970-01-01 are used as the x-axis for the
    regressions. Such values can be accessed with
    ``self.time_points.values.astype("datetime64[s]").astype("int")``. This
    decision does not matter for the gradients, but is important for the
    intercept values.

    Returns
    -------
    list of dict[str : Any]
        List of dictionaries. Each dictionary contains the metadata for the
        timeseries plus the gradient (with key ``"gradient"``) and intercept (
        with key ``"intercept"``). The gradient and intercept are stored as
        :obj:`pint.Quantity`.
    """
    _, _, time_unit, gradients, intercepts, meta = _calculate_linear_regression(self)

    out = []
    for row_meta, gradient, intercept in zip(
        meta.to_dict("records"), gradients, intercepts,
    ):
        unit = row_meta.pop("unit")

        row_meta["gradient"] = gradient * unit_registry(
            "{} / {}".format(unit, time_unit)
        )
        row_meta["intercept"] = intercept * unit_registry(unit)

        out.append(row_meta)

    return out


def _convert_linear_regression_raw_to_pdf(raw, key_to_keep, unit):
    pdf_dicts = []
    for r in raw:
        transformed = {
            k: v for k, v in r.items() if k not in ["gradient", "intercept", "unit"]
        }
        if unit is None:
            transformed[key_to_keep] = r[key_to_keep].magnitude
            transformed["unit"] = str(r["gradient"].units)
        else:
            transformed[key_to_keep] = r[key_to_keep].to(unit).magnitude
            transformed["unit"] = unit

        pdf_dicts.append(transformed)

    return pd.DataFrame(pdf_dicts)


def linear_regression_gradient(self, unit=None):
    """
    Calculate gradients of a linear regression of each timeseries

    Parameters
    ----------
    unit : str
        Output unit for gradients. If not supplied, the gradients' units will
        not be converted to a common unit.

    Returns
    -------
    :obj:`pd.DataFrame`
        ``self.meta`` plus a column with the value of the gradient for each
        timeseries. The ``"unit"`` column is updated to show the unit of the
        gradient.
    """
    raw = self.linear_regression()

    return _convert_linear_regression_raw_to_pdf(raw, "gradient", unit)


def linear_regression_intercept(self, unit=None):
    """
    Calculate intercepts of a linear regression of each timeseries

    Note
    ----
    Times in seconds since 1970-01-01 are used as the x-axis for the
    regressions. Such values can be accessed with
    ``self.time_points.values.astype("datetime64[s]").astype("int")``. This
    decision does not matter for the gradients, but is important for the
    intercept values.

    Parameters
    ----------
    unit : str
        Output unit for gradients. If not supplied, the gradients' units will
        not be converted to a common unit.

    Returns
    -------
    :obj:`pd.DataFrame`
        ``self.meta`` plus a column with the value of the gradient for each
        timeseries. The ``"unit"`` column is updated to show the unit of the
        gradient.
    """
    raw = self.linear_regression()

    return _convert_linear_regression_raw_to_pdf(raw, "intercept", unit)


def linear_regression_scmrun(self):
    """
    Re-calculate the timeseries based on a linear regression

    Returns
    -------
    :obj:`scmdata.ScmRun`
        The timeseries, re-calculated based on a linear regression
    """
    (
        times_numpy,
        times_in_s,
        time_unit,
        gradients,
        intercepts,
        meta,
    ) = _calculate_linear_regression(self)

    out_shape = (meta.shape[0], len(times_in_s))
    regression_timeseries = (
        np.broadcast_to(gradients, out_shape[::-1]).T * times_in_s
        + intercepts[:, np.newaxis]
    )

    out = type(self)(
        data=regression_timeseries.T,
        index=times_numpy,
        columns=meta.to_dict(orient="list"),
    )

    return out


def _calculate_linear_regression(in_scmrun):
    time_unit = "s"
    times_numpy = in_scmrun.time_points.values.astype(
        "datetime64[{}]".format(time_unit)
    )
    times_in_s = times_numpy.astype("int")

    ts = in_scmrun.timeseries()
    if ts.isnull().sum().sum() > 0:
        warnings.warn(
            "You are calculating a linear regression of data which contains "
            "nans so your result will also contain nans. Perhaps you want to "
            "remove the nans before calculating the regression using a "
            "combination of :meth:`filter` and :meth:`interpolate`?"
        )

    res = np.polyfit(times_in_s, ts.T, 1)
    gradients = res[0, :]
    intercepts = res[1, :]

    meta = ts.index.to_frame().reset_index(drop=True)

    return times_numpy, times_in_s, time_unit, gradients, intercepts, meta


def inject_ops_methods(cls):
    """
    Inject the operation methods

    Parameters
    ----------
    cls
        Target class
    """
    methods = [
        ("subtract", subtract),
        ("add", add),
        ("multiply", multiply),
        ("divide", divide),
        ("integrate", integrate),
        ("delta_per_delta_time", delta_per_delta_time),
        ("linear_regression", linear_regression),
        ("linear_regression_gradient", linear_regression_gradient),
        ("linear_regression_intercept", linear_regression_intercept),
        ("linear_regression_scmrun", linear_regression_scmrun),
    ]

    for name, f in methods:
        setattr(cls, name, f)



"""
Time period handling and interpolation

A large portion of this module was originally from openscm. Thanks to the original author, Sven Willner
"""

from datetime import datetime

import cftime
import numpy as np
import pandas as pd
from dateutil import parser

try:
    import scipy.interpolate

    has_scipy = True
except ImportError:  # pragma: no cover
    scipy = None
    has_scipy = False


_TARGET_TYPE = np.int64
_TARGET_DTYPE = "datetime64[s]"


class InsufficientDataError(Exception):
    """
    Insufficient data is available to interpolate/extrapolate
    """

    pass


def _float_year_to_datetime(inp: float) -> np.datetime64:
    year = int(inp)
    fractional_part = inp - year
    return np.datetime64(  # pylint: disable=too-many-function-args
        year - 1970, "Y"
    ) + np.timedelta64(  # pylint: disable=too-many-function-args
        int(
            (datetime(year + 1, 1, 1) - datetime(year, 1, 1)).total_seconds()
            * fractional_part
        ),
        "s",
    )


_ufunc_float_year_to_datetime = np.frompyfunc(_float_year_to_datetime, 1, 1)
_ufunc_str_to_datetime = np.frompyfunc(parser.parse, 1, 1)


def _parse_datetime(inp: np.ndarray) -> np.ndarray:
    try:
        return _ufunc_float_year_to_datetime(inp.astype(float))
    except (TypeError, ValueError):
        return _ufunc_str_to_datetime(inp)


def _format_datetime(dts: np.ndarray) -> np.ndarray:
    """
    Convert an array to an array of :class:`np.datetime64`.

    Parameters
    ----------
    dts
        Input to attempt to convert

    Returns
    -------
    :class:`np.ndarray` of :class:`np.datetime64`
        Converted array

    Raises
    ------
    ValueError
        If one of the values in :obj:`dts` cannot be converted to :class:`np.datetime64`
    """
    if len(dts) <= 0:  # pylint: disable=len-as-condition
        return np.array([], dtype=_TARGET_DTYPE)

    dtype = dts.dtype.type
    if dts.dtype.kind == "O":
        dtype = np.dtype(type(dts[0])).type
    if issubclass(dtype, np.datetime64):
        return np.asarray(dts, dtype=_TARGET_DTYPE)
    if issubclass(dtype, np.floating):
        return _ufunc_float_year_to_datetime(dts).astype(_TARGET_DTYPE)
    if issubclass(dtype, np.integer):
        return (np.asarray(dts) - 1970).astype("datetime64[Y]").astype(_TARGET_DTYPE)
    if issubclass(dtype, str):
        return _parse_datetime(dts).astype(_TARGET_DTYPE)
    return np.asarray(dts, dtype=_TARGET_DTYPE)


class TimePoints:
    """
    Handles time points by wrapping :class:`np.ndarray` of :class:`np.datetime64`..
    """

    def __init__(self, values):
        """
        Initialize.

        Parameters
        ----------
        values
            Time points array to handle
        """
        self._values = _format_datetime(np.asarray(values))

    def __len__(self) -> int:
        """
        Get the number of time points.
        """
        return len(self._values)

    @property
    def values(self) -> np.ndarray:
        """
        Time points
        """
        return self._values

    def to_index(self) -> pd.Index:
        """
        Get time points as :class:`pd.Index`.

        Returns
        -------
        :class:`pd.Index`
            :class:`pd.Index` of :class:`np.dtype` :class:`object` with name ``"time"``
            made from the time points represented as :class:`datetime.datetime`.
        """
        return pd.Index(self._values.astype(object), dtype=object, name="time")

    def as_cftime(self) -> list:
        """
        Get as cftime datetimes

        Returns
        -------
        list of cftime.datetime
        """
        return [
            cftime.datetime(*dt.timetuple()[:6]) for dt in self._values.astype(object)
        ]

    def years(self) -> np.ndarray:
        """
        Get year of each time point.

        Returns
        -------
        :obj:`np.array` of :obj:`int`
            Year of each time point
        """
        return np.vectorize(getattr)(self._values.astype(object), "year")

    def months(self) -> np.ndarray:
        """
        Get month of each time point.

        Returns
        -------
        :obj:`np.array` of :obj:`int`
            Month of each time point
        """
        return np.vectorize(getattr)(self._values.astype(object), "month")

    def days(self) -> np.ndarray:
        """
        Get day of each time point.

        Returns
        -------
        :obj:`np.array` of :obj:`int`
            Day of each time point
        """
        return np.vectorize(getattr)(self._values.astype(object), "day")

    def hours(self) -> np.ndarray:
        """
        Get hour of each time point.

        Returns
        -------
        :obj:`np.array` of :obj:`int`
            Hour of each time point
        """
        return np.vectorize(getattr)(self._values.astype(object), "hour")

    def weekdays(self) -> np.ndarray:
        """
        Get weekday of each time point.

        Returns
        -------
        :obj:`np.array` of :obj:`int`
            Day of the week of each time point
        """
        return np.vectorize(datetime.weekday)(self._values.astype(object))


class TimeseriesConverter:
    """
    Interpolator used to convert data between different time bases

    This is a modified version originally in :mod:`openscm.time.TimeseriesConverter`.
    The integral preserving interpolation was removed as it is outside the scope of
    this package.

    Parameters
    ----------
    source_time_points: np.ndarray
        Source timeseries time points
    target_time_points: np.ndarray
        Target timeseries time points
    interpolation_type : {"linear"}
        Interpolation type. Options are 'linear'
    extrapolation_type : {"linear", "constant", None}
        Extrapolation type. Options are None, 'linear' or 'constant'

    Raises
    ------
    InsufficientDataError
        Timeseries too short to extrapolate
    """

    def __init__(
        self,
        source_time_points: np.ndarray,
        target_time_points: np.ndarray,
        interpolation_type="linear",
        extrapolation_type="linear",
    ):

        self.source = (
            np.array(source_time_points)
            .astype(_TARGET_DTYPE)
            .astype(_TARGET_TYPE, copy=True)
        )
        self.target = (
            np.array(target_time_points)
            .astype(_TARGET_DTYPE)
            .astype(_TARGET_TYPE, copy=True)
        )
        self.interpolation_type = interpolation_type
        self.extrapolation_type = extrapolation_type

        if not self.points_are_compatible(self.source, self.target):
            error_msg = (
                "Target time points are outside the source time points, use an "
                "extrapolation type other than None"
            )
            raise InsufficientDataError(error_msg)

    def points_are_compatible(self, source: np.ndarray, target: np.ndarray) -> bool:
        """
        Are the two sets of time points compatible i.e. can I convert between the two?

        Parameters
        ----------
        source
            Source timeseries time points
        target
            Target timeseries time points

        Returns
        -------
        bool
            Can I convert between the time points?
        """
        if self.extrapolation_type is None and (
            source[0] > target[0] or source[-1] < target[-1]
        ):
            return False

        return True

    def _get_scipy_extrapolation_args(self, values: np.ndarray):
        if self.extrapolation_type == "linear":
            return {"fill_value": "extrapolate"}
        if self.extrapolation_type == "constant":
            return {"fill_value": (values[0], values[-1]), "bounds_error": False}
        # TODO: add cubic support
        return {}

    def _get_scipy_interpolation_arg(self) -> str:
        if self.interpolation_type == "linear":
            return "linear"
        # TODO: add cubic support
        raise NotImplementedError

    def _convert(
        self,
        values: np.ndarray,
        source_time_points: np.ndarray,
        target_time_points: np.ndarray,
    ) -> np.ndarray:
        """
        Wrap :func:`_convert_unsafe` to provide proper error handling.

        Any nan values are removed from :obj:`source` before interpolation

        Parameters
        ----------
        values
            Array of data to convert
        source_time_points
            Source timeseries time points
        target_time_points
            Target timeseries time points

        Raises
        ------
        InsufficientDataError
            Length of the time series is too short to convert
        InsufficientDataError
            Target time points are outside the source time points and
            :attr:`extrapolation_type` is 'NONE'
        ImportError
            Optional dependency scipy has not been installed

        Returns
        -------
        np.ndarray
            Converted time period average data for timeseries :obj:`values`
        """
        values = np.asarray(values)
        # Check for nans
        nan_mask = np.isnan(values)
        if nan_mask.sum():
            values = values[~nan_mask]
            source_time_points = source_time_points[~nan_mask]

        if len(values) < 3:
            raise InsufficientDataError

        try:
            return self._convert_unsafe(values, source_time_points, target_time_points)
        except Exception:  # pragma: no cover # emergency valve
            print("numpy interpolation failed...")
            raise

    def _convert_unsafe(
        self,
        values: np.ndarray,
        source_time_points: np.ndarray,
        target_time_points: np.ndarray,
    ) -> np.ndarray:
        if not has_scipy:
            raise ImportError("scipy is not installed. Run 'pip install scipy'")

        res_point = scipy.interpolate.interp1d(
            source_time_points.astype(_TARGET_TYPE),
            values,
            kind=self._get_scipy_interpolation_arg(),
            **self._get_scipy_extrapolation_args(values),
        )

        return res_point(target_time_points.astype(_TARGET_TYPE))

    def convert_from(self, values: np.ndarray) -> np.ndarray:
        """
        Convert value **from** source timeseries time points to target timeseries time
        points.

        Parameters
        ----------
        values: np.ndarray
            Value

        Returns
        -------
        np.ndarray
            Converted data for timeseries :obj:`values` into the target timebase
        """
        return self._convert(values, self.source, self.target)

    def convert_to(self, values: np.ndarray) -> np.ndarray:
        """
        Convert value from target timeseries time points **to** source timeseries time
        points.

        Parameters
        ----------
        values: np.ndarray
            Value

        Returns
        -------
        np.ndarray
            Converted data for timeseries :obj:`values` into the source timebase
        """
        return self._convert(values, self.target, self.source)



"""
Custom errors and exceptions used by scmdata
"""
import pandas as pd


class NonUniqueMetadataError(ValueError):
    """
    Error raised when metadata is not unique
    """

    def __init__(self, meta):
        # format table to show the metadata clash
        dup = meta.astype(str).groupby(meta.columns.tolist(), as_index=False).size()
        if isinstance(dup, pd.Series):
            # pandas < 1.1 Groupby.size returns a series
            dup.name = "repeats"
            dup = dup.to_frame().reset_index()
        else:
            dup = dup.rename(columns={"size": "repeats"})
        dup = dup[dup.repeats > 1]
        msg = (
            "Duplicate metadata (numbers show how many times the given "
            "metadata is repeated).\n{}".format(dup)
        )

        super().__init__(msg)


class MissingRequiredColumnError(ValueError):
    """
    Error raised when an operation produces missing metadata columns
    """

    def __init__(self, columns):
        self.columns = columns
        msg = "missing required columns `{}`!".format(columns)

        super().__init__(msg)



"""
scmdata, simple data handling for simple climate model data
"""
from ._version import get_versions  # isort:skip

__version__ = get_versions()["version"]
del get_versions


from scmdata.run import ScmRun, run_append  # noqa: F401, E402

# TODO: remove in next major release
REQUIRED_COLS = ScmRun.required_cols  # noqa



"""
TimeSeries handling

Functionality for handling and storing individual time-series
"""

import copy
import datetime as dt
import functools
from typing import Any, Callable, List, Union

import numpy as np
import pint
import xarray as xr
from openscm_units import unit_registry as ur
from xarray.core.ops import inject_binary_ops

from .time import TimePoints, TimeseriesConverter


class _Counter:
    def __init__(self):
        self.count = 0

    def __call__(self):
        val = self.count
        self.count += 1

        return val

    def reset(self):
        self.count = 0


get_default_name = _Counter()


class TimeSeries:
    """
    A 1D time-series with metadata

    Proxies an xarray.DataArray with a single time dimension
    """

    def __init__(self, data, time=None, **kwargs):
        """
        Initialise a :obj:`TimeSeries` instance

        Parameters
        ----------
        data : array_like
            Data to be held by the :obj:`TimeSeries` instance. ``data`` must
            be one-dimensional. If ``data`` is an :obj:`xr.DataArray`
            instance, its single dimension must be ``"time"``. If ``data``
            is not an :obj:`xr.DataArray`, then ``time`` must also be supplied.

        time : array_like or None
            Only used if ``data`` is not an :obj:`xr.DataArray`. These become
            the time axis of ``self._data``.

        **kwargs
            Only used if `data`` is not an :obj:`xr.DataArray`. Passed to the
            :obj:`xr.DataArray` constructor.

        Raises
        ------
        ValueError
            ``data`` is not one-dimensional

        TypeError
            ``data`` is an :obj:`xr.DataArray` and ``time is not None``

        ValueError
            ``data`` is an :obj:`xr.DataArray` and its dimension is not named
            ``"time"``.

        TypeError
            ``data`` is not an :obj:`xr.DataArray` and ``time is None``

        ValueError
            ``data`` is not an :obj:`xr.DataArray` and ``coords`` is supplied
            via ``**kwargs``
        """
        values = np.asarray(data)

        if values.ndim != 1:
            raise ValueError("data must be 1d")

        if isinstance(data, xr.DataArray):
            if time is not None:
                raise TypeError(
                    "If data is an :obj:`xr.DataArray` instance, time must be " "`None`"
                )

            if data.dims != ("time",):
                raise ValueError(
                    "If data is an :obj:`xr.DataArray` instance, its only "
                    "dimension must be named `'time'`"
                )
            self._data = data

        else:
            if time is None:
                raise TypeError(
                    "If data is not an :obj:`xr.DataArray` instance, `time` "
                    "must not be `None`"
                )

            if "coords" in kwargs:
                raise ValueError(
                    "If ``data`` is not an :obj:`xr.DataArray`, `coords` must "
                    "not be supplied via `kwargs` because it will be "
                    "automatically filled with the value of `time`."
                )

            # Auto incrementing name
            if "name" not in kwargs:
                kwargs["name"] = get_default_name()

            if isinstance(time, tuple):
                time = list(time)

            self._data = xr.DataArray(values, coords=[("time", time)], **kwargs)

    def __repr__(self):
        return self._data.__repr__()

    @property
    def name(self):
        """
        Timeseries name

        If no name was provided this will be an automatically incrementing number
        """
        return self._data.name

    def __len__(self):
        """
        Length of the time-series (number of time points)

        Returns
        -------
        int
        """
        return len(self._data["time"])

    def copy(self):
        """
        Create a deep copy of the timeseries.

        Any further modifications to the :obj:`Timeseries` returned copy will not be reflecting
        in the current :obj:`Timeseries`

        Returns
        -------
        :obj:`Timeseries`
        """
        return copy.deepcopy(self)

    @property
    def meta(self):
        """
        Metadata associated with the timeseries

        Returns
        -------
        dict
        """
        return self._data.attrs

    @property
    def time_points(self):
        """
        Time points of the data

        Returns
        -------
        :obj:`np.ndarray`
        """
        return TimePoints(self._data.coords["time"].values)

    @property
    def values(self):
        """
        Get the data as a numpy array

        Returns
        -------
        :obj:`np.ndarray`
        """
        return self._data.values

    def __getitem__(self, item):
        res = self._data.__getitem__(item)
        if res.ndim == 0:
            return res
        return TimeSeries(res)

    def __setitem__(self, key, value):
        self._data.__setitem__(key, value)

    @staticmethod
    def _binary_op(
        f: Callable[..., Any], reflexive=False, **kwargs,
    ) -> Callable[..., "TimeSeries"]:
        @functools.wraps(f)
        def func(self, other):
            other_data = getattr(other, "_data", other)

            if isinstance(other, pint.Quantity):
                try:
                    self_data = self._data * ur(self.meta["unit"])
                except KeyError:
                    # let Pint assume dimensionless and raise an error as
                    # necessary
                    self_data = self._data
            else:
                self_data = self._data

            ts = f(self_data, other_data) if not reflexive else f(other_data, self_data)
            ts.attrs = self._data.attrs
            if isinstance(other, pint.Quantity):
                ts.attrs["unit"] = str(ts.data.units)
                ts.data = ts.data.magnitude

            return TimeSeries(ts)

        return func

    @staticmethod
    def _inplace_binary_op(f: Callable) -> Callable[..., "TimeSeries"]:
        @functools.wraps(f)
        def func(self, other):
            other_data = getattr(other, "_data", other)
            f(self._data, other_data)
            return self

        return func

    def reindex(self, time, **kwargs):
        """
        Update the time dimension, filling in the missing values with NaN's

        This is different to interpolating to fill in the missing values. Uses
        `xarray.DataArray.reindex` to perform the reindexing

        Parameters
        ----------
        time : `obj`:np.ndarray
            Time values to reindex the data to. Should be ``np.datetime64``
            values

        **kwargs
            Additional arguments passed to xarray's DataArray.reindex function

        Returns
        -------
        :obj:`TimeSeries`
            A new TimeSeries with the new time dimension

        References
        ----------
        http://xarray.pydata.org/en/stable/generated/xarray.DataArray.reindex_like.html#xarray.DataArray.reindex_like
        """
        return TimeSeries(self._data.reindex({"time": time}, **kwargs))

    def interpolate(
        self,
        target_times: Union[np.ndarray, List[Union[dt.datetime, int]]],
        interpolation_type: str = "linear",
        extrapolation_type: str = "linear",
    ):
        """
        Interpolate the timeseries onto a new time axis

        Parameters
        ----------
        target_times
            Time grid onto which to interpolate
        interpolation_type: str
            Interpolation type. Options are 'linear'
        extrapolation_type: str or None
            Extrapolation type. Options are None, 'linear' or 'constant'
        Returns
        -------
        :obj:`TimeSeries`
            A new TimeSeries with the new time dimension
        """
        target_times = TimePoints(target_times)
        timeseries_converter = TimeseriesConverter(
            self.time_points.values,
            target_times.values,
            interpolation_type=interpolation_type,
            extrapolation_type=extrapolation_type,
        )

        ts = self.reindex(target_times.as_cftime())
        ts._data[:] = timeseries_converter.convert_from(self._data.values)

        return ts


inject_binary_ops(TimeSeries)
